import time
import logger
import random
import string
import copy
import json
import re
from threading import Thread
from memcacheConstants import ERR_NOT_FOUND
from membase.api.rest_client import RestConnection, Bucket
from membase.api.exception import BucketCreationException
from membase.helper.bucket_helper import BucketOperationHelper
from memcached.helper.data_helper import KVStoreAwareSmartClient, VBucketAwareMemcached, MemcachedClientHelper
from couchbase.document import DesignDocument, View
from mc_bin_client import MemcachedError
from tasks.future import Future
from membase.api.exception import DesignDocCreationException, QueryViewException, ReadDocumentException, RebalanceFailedException, \
                                        GetBucketInfoFailed, CompactViewFailed, SetViewInfoNotFound, FailoverFailedException

#TODO: Setup stacktracer
#TODO: Needs "easy_install pygments"
#import stacktracer
#stacktracer.trace_start("trace.html",interval=30,auto=True) # Set auto flag to always update file!


PENDING = 'PENDING'
EXECUTING = 'EXECUTING'
CHECKING = 'CHECKING'
FINISHED = 'FINISHED'

class Task(Future):
    def __init__(self, name):
        Future.__init__(self)
        self.log = logger.Logger.get_logger()
        self.state = PENDING
        self.name = name
        self.cancelled = False
        self.retries = 0
        self.res = None

    def step(self, task_manager):
        if not self.done():
            if self.state == PENDING:
                self.state = EXECUTING
                task_manager.schedule(self)
            elif self.state == EXECUTING:
                self.execute(task_manager)
            elif self.state == CHECKING:
                self.check(task_manager)
            elif self.state != FINISHED:
                raise Exception("Bad State in {0}: {1}".format(self.name, self.state))

    def execute(self, task_manager):
        raise NotImplementedError

    def check(self, task_manager):
        raise NotImplementedError

class NodeInitializeTask(Task):
    def __init__(self, server):
        Task.__init__(self, "node_init_task")
        self.server = server
        self.quota = 0

    def execute(self, task_manager):
        rest = RestConnection(self.server)
        username = self.server.rest_username
        password = self.server.rest_password
        rest.init_cluster(username, password)
        info = rest.get_nodes_self()
        if info is None:
            self.state = FINISHED
            self.set_exception(Exception('unable to get information on a server %s, it is available?' % (self.server.ip)))
        self.quota = int(info.mcdMemoryReserved * 2 / 3)
        rest.init_cluster_memoryQuota(username, password, self.quota)
        self.state = CHECKING
        task_manager.schedule(self)

    def check(self, task_manager):
        self.state = FINISHED
        self.set_result(self.quota)


class BucketCreateTask(Task):
    def __init__(self, server, bucket='default', replicas=1, size=0, port=11211,
                 password=None):
        Task.__init__(self, "bucket_create_task")
        self.server = server
        self.bucket = bucket
        self.replicas = replicas
        self.port = port
        self.size = size
        self.password = password

    def execute(self, task_manager):
        rest = RestConnection(self.server)
        if self.size <= 0:
            info = rest.get_nodes_self()
            self.size = info.memoryQuota * 2 / 3

        authType = 'none' if self.password is None else 'sasl'

        try:
            rest.create_bucket(bucket=self.bucket,
                               ramQuotaMB=self.size,
                               replicaNumber=self.replicas,
                               proxyPort=self.port,
                               authType=authType,
                               saslPassword=self.password)
            self.state = CHECKING
            task_manager.schedule(self)
        except BucketCreationException as e:
            self.state = FINISHED
            self.set_exception(e)
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    def check(self, task_manager):
        try:
            if BucketOperationHelper.wait_for_memcached(self.server, self.bucket):
                self.log.info("bucket '{0}' was created with per node RAM quota: {1}".format(self.bucket, self.size))
                self.set_result(True)
                self.state = FINISHED
                return
            else:
                self.log.info("vbucket map not ready after try {0}".format(self.retries))
        except Exception:
            self.log.info("vbucket map not ready after try {0}".format(self.retries))
        self.retries = self.retries + 1
        task_manager.schedule(self)


class BucketDeleteTask(Task):
    def __init__(self, server, bucket="default"):
        Task.__init__(self, "bucket_delete_task")
        self.server = server
        self.bucket = bucket

    def execute(self, task_manager):
        rest = RestConnection(self.server)
        try:
            if rest.delete_bucket(self.bucket):
                self.state = CHECKING
                task_manager.schedule(self)
            else:
                self.state = FINISHED
                self.set_result(False)
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)


    def check(self, task_manager):
        rest = RestConnection(self.server)
        try:
            if BucketOperationHelper.wait_for_bucket_deletion(self.bucket, rest, 200):
                self.set_result(True)
            else:
                self.set_result(False)
            self.state = FINISHED
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

class RebalanceTask(Task):
    def __init__(self, servers, to_add=[], to_remove=[], do_stop=False, progress=30):
        Task.__init__(self, "rebalance_task")
        self.servers = servers
        self.to_add = to_add
        self.to_remove = to_remove

    def execute(self, task_manager):
        try:
            self.add_nodes(task_manager)
            self.start_rebalance(task_manager)
            self.state = CHECKING
            task_manager.schedule(self)
        except Exception as e:
            self.state = FINISHED
            self.set_exception(e)

    def add_nodes(self, task_manager):
        master = self.servers[0]
        rest = RestConnection(master)
        for node in self.to_add:
            self.log.info("adding node {0}:{1} to cluster".format(node.ip, node.port))
            rest.add_node(master.rest_username, master.rest_password,
                          node.ip, node.port)

    def start_rebalance(self, task_manager):
        rest = RestConnection(self.servers[0])
        nodes = rest.node_statuses()
        ejectedNodes = []
        for server in self.to_remove:
            for node in nodes:
                if server.ip == node.ip and int(server.port) == int(node.port):
                    ejectedNodes.append(node.id)
        rest.rebalance(otpNodes=[node.id for node in nodes], ejectedNodes=ejectedNodes)

    def check(self, task_manager):
        rest = RestConnection(self.servers[0])
        progress = -100
        try:
            progress = rest._rebalance_progress()
        except RebalanceFailedException as ex:
            self.state = FINISHED
            self.set_exception(ex)
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)
        if progress != -1 and progress != 100:
            task_manager.schedule(self, 10)
        else:
            self.log.info("rebalancing was completed with progress: {0}%".format(progress))
            self.state = FINISHED
            self.set_result(True)

class StatsWaitTask(Task):
    EQUAL = '=='
    NOT_EQUAL = '!='
    LESS_THAN = '<'
    LESS_THAN_EQ = '<='
    GREATER_THAN = '>'
    GREATER_THAN_EQ = '>='

    def __init__(self, servers, bucket, param, stat, comparison, value):
        Task.__init__(self, "stats_wait_task")
        self.servers = servers
        self.bucket = bucket
        if isinstance(bucket, Bucket):
            self.bucket = bucket.name
        self.param = param
        self.stat = stat
        self.comparison = comparison
        self.value = value
        self.conns = {}

    def execute(self, task_manager):
        self.state = CHECKING
        task_manager.schedule(self)

    def check(self, task_manager):
        stat_result = 0
        for server in self.servers:
            client = self._get_connection(server)
            try:
                stats = client.stats(self.param)
                if not stats.has_key(self.stat):
                    self.state = FINISHED
                    self.set_exception(Exception("Stat {0} not found".format(self.stat)))
                    return
                if stats[self.stat].isdigit():
                    stat_result += long(stats[self.stat])
                else:
                    stat_result = stats[self.stat]
            except EOFError as ex:
                self.state = FINISHED
                self.set_exception(ex)

        if not self._compare(self.comparison, str(stat_result), self.value):
            self.log.info("Not Ready: %s %s %s %s expected on %s" % (self.stat, stat_result,
                      self.comparison, self.value, self._stringify_servers()))
            task_manager.schedule(self, 5)
            return
        self.log.info("Saw %s %s %s %s expected on %s" % (self.stat, stat_result,
                      self.comparison, self.value, self._stringify_servers()))

        for server, conn in self.conns.items():
            conn.close()
        self.state = FINISHED
        self.set_result(True)

    def _stringify_servers(self):
        return ''.join([`server.ip` for server in self.servers])

    def _get_connection(self, server):
        if not self.conns.has_key(server):
            self.conns[server] = MemcachedClientHelper.direct_client(server, self.bucket)
        return self.conns[server]

    def _compare(self, cmp_type, a, b):
        if isinstance(b, (int, long)) and a.isdigit():
            a = long(a)
        elif isinstance(b, (int, long)) and not a.isdigit():
                return False
        if (cmp_type == StatsWaitTask.EQUAL and a == b) or\
            (cmp_type == StatsWaitTask.NOT_EQUAL and a != b) or\
            (cmp_type == StatsWaitTask.LESS_THAN_EQ and a <= b) or\
            (cmp_type == StatsWaitTask.GREATER_THAN_EQ and a >= b) or\
            (cmp_type == StatsWaitTask.LESS_THAN and a < b) or\
            (cmp_type == StatsWaitTask.GREATER_THAN and a > b):
            return True
        return False

class GenericLoadingTask(Thread, Task):
    def __init__(self, server, bucket, kv_store):
        Thread.__init__(self)
        Task.__init__(self, "load_gen_task")
        self.kv_store = kv_store
        self.client = VBucketAwareMemcached(RestConnection(server), bucket)

    def execute(self, task_manager):
        self.start()
        self.state = EXECUTING

    def check(self, task_manager):
        pass

    def run(self):
        while self.has_next() and not self.done():
            self.next()
        self.state = FINISHED
        self.set_result(True)

    def has_next(self):
        raise NotImplementedError

    def next(self):
        raise NotImplementedError

    def _unlocked_create(self, partition, key, value):
        try:
            value_json = json.loads(value)
            value_json['mutated'] = 0
            value = json.dumps(value_json)
        except ValueError:
            index = random.choice(range(len(value)))
            value = value[0:index] + random.choice(string.ascii_uppercase) + value[index + 1:]

        try:
            self.client.set(key, self.exp, 0, value)
            partition.set(key, value, self.exp)
        except MemcachedError as error:
            self.state = FINISHED
            self.set_exception(error)

    def _unlocked_read(self, partition, key):
        try:
            o, c, d = self.client.get(key)
        except MemcachedError as error:
            if error.status == ERR_NOT_FOUND and partition.get_valid(key) is None:
                pass
            else:
                self.state = FINISHED
                self.set_exception(error)

    def _unlocked_update(self, partition, key):
        value = partition.get_valid(key)
        if value is None:
            return
        try:
            value_json = json.loads(value)
            value_json['mutated'] += 1
            value = json.dumps(value_json)
        except ValueError:
            index = random.choice(range(len(value)))
            value = value[0:index] + random.choice(string.ascii_uppercase) + value[index + 1:]

        try:
            self.client.set(key, self.exp, 0, value)
            partition.set(key, value, self.exp)
        except MemcachedError as error:
            self.state = FINISHED
            self.set_exception(error)

    def _unlocked_delete(self, partition, key):
        try:
            self.client.delete(key)
            partition.delete(key)
        except MemcachedError as error:
            if error.status == ERR_NOT_FOUND and partition.get_valid(key) is None:
                pass
            else:
                self.state = FINISHED
                self.set_exception(error)

class LoadDocumentsTask(GenericLoadingTask):
    def __init__(self, server, bucket, generator, kv_store, op_type, exp):
        GenericLoadingTask.__init__(self, server, bucket, kv_store)
        self.generator = generator
        self.op_type = op_type
        self.exp = exp

    def has_next(self):
        return self.generator.has_next()

    def next(self):
        key, value = self.generator.next()
        partition = self.kv_store.acquire_partition(key)
        if self.op_type == 'create':
            self._unlocked_create(partition, key, value)
        elif self.op_type == 'read':
            self._unlocked_read(partition, key)
        elif self.op_type == 'update':
            self._unlocked_update(partition, key)
        elif self.op_type == 'delete':
            self._unlocked_delete(partition, key)
        else:
            self.state = FINISHED
            self.set_exception(Exception("Bad operation type: %s" % self.op_type))
        self.kv_store.release_partition(key)

class WorkloadTask(GenericLoadingTask):
    def __init__(self, server, bucket, kv_store, num_ops, create, read, update, delete, exp):
        GenericLoadingTask.__init__(self, server, bucket, kv_store)
        self.itr = 0
        self.num_ops = num_ops
        self.create = create
        self.read = create + read
        self.update = create + read + update
        self.delete = create + read + update + delete
        self.exp = exp

    def has_next(self):
        if self.num_ops == 0 or self.itr < self.num_ops:
            return True
        return False

    def next(self):
        self.itr += 1
        rand = random.randint(1, self.delete)
        if rand > 0 and rand <= self.create:
            self._create_random_key()
        elif rand > self.create and rand <= self.read:
            self._get_random_key()
        elif rand > self.read and rand <= self.update:
            self._update_random_key()
        elif rand > self.update and rand <= self.delete:
            self._delete_random_key()

    def _get_random_key(self):
        partition, part_num = self.kv_store.acquire_random_partition()
        if partition is None:
            return

        key = partition.get_random_valid_key()
        if key is None:
            self.kv_store.release_partition(part_num)
            return

        self._unlocked_read(partition, key)
        self.kv_store.release_partition(part_num)

    def _create_random_key(self):
        partition, part_num = self.kv_store.acquire_random_partition(False)
        if partition is None:
            return

        key = partition.get_random_deleted_key()
        if key is None:
            self.kv_store.release_partition(part_num)
            return

        value = partition.get_deleted(key)
        if value is None:
            self.kv_store.release_partition(part_num)
            return

        self._unlocked_create(partition, key, value)
        self.kv_store.release_partition(part_num)

    def _update_random_key(self):
        partition, part_num = self.kv_store.acquire_random_partition()
        if partition is None:
            return

        key = partition.get_random_valid_key()
        if key is None:
            self.kv_store.release_partition(part_num)
            return

        self._unlocked_update(partition, key)
        self.kv_store.release_partition(part_num)

    def _delete_random_key(self):
        partition, part_num = self.kv_store.acquire_random_partition()
        if partition is None:
            return

        key = partition.get_random_valid_key()
        if key is None:
            self.kv_store.release_partition(part_num)
            return

        self._unlocked_delete(partition, key)
        self.kv_store.release_partition(part_num)

class ValidateDataTask(GenericLoadingTask):
    def __init__(self, server, bucket, kv_store):
        GenericLoadingTask.__init__(self, server, bucket, kv_store)
        self.valid_keys, self.deleted_keys = kv_store.key_set()
        self.num_valid_keys = len(self.valid_keys)
        self.num_deleted_keys = len(self.deleted_keys)
        self.itr = 0

    def has_next(self):
        if self.itr < (self.num_valid_keys + self.num_deleted_keys):
            return True
        return False

    def next(self):
        if self.itr < self.num_valid_keys:
            self._check_valid_key(self.valid_keys[self.itr])
        else:
            self._check_deleted_key(self.deleted_keys[self.itr - self.num_valid_keys])
        self.itr += 1

    def _check_valid_key(self, key):
        partition = self.kv_store.acquire_partition(key)

        value = partition.get_valid(key)
        if value is None:
            self.kv_store.release_partition(key)
            return
        value = json.dumps(value)

        try:
            o, c, d = self.client.get(key)
            if d != json.loads(value):
                self.state = FINISHED
                self.set_exception(Exception('Bad result: %s != %s' % (json.dumps(d), value)))
        except MemcachedError as error:
            if error.status == ERR_NOT_FOUND and partition.get_valid(key) is None:
                pass
            else:
                self.state = FINISHED
                self.set_exception(error)
        self.kv_store.release_partition(key)

    def _check_deleted_key(self, key):
        partition = self.kv_store.acquire_partition(key)

        try:
            o, c, d = self.client.delete(key)
            if partition.get_valid(key) is not None:
                self.state = FINISHED
                self.set_exception(Exception('Not Deletes: %s' % (key)))
        except MemcachedError as error:
            if error.status == ERR_NOT_FOUND:
                pass
            else:
                self.state = FINISHED
                self.set_exception(error)
        self.kv_store.release_partition(key)


class VerifyRevIdTask(GenericLoadingTask):
    def __init__(self, src_server, dest_server, bucket, kv_store, ops_perf):
        GenericLoadingTask.__init__(self, src_server, bucket, kv_store)
        self.client_dest = VBucketAwareMemcached(RestConnection(dest_server), bucket)

        self.valid_keys, self.deleted_keys = kv_store.key_set()
        self.num_valid_keys = len(self.valid_keys)
        self.num_deleted_keys = len(self.deleted_keys)
        self.itr = 0
        self.err_count = 0
        self.ops_perf = ops_perf

    def has_next(self):
        if self.ops_perf == "delete":
            if self.itr < self.num_deleted_keys:
                return True
            return False
        elif self.ops_perf == "update":
            if self.itr < self.num_valid_keys:
                return True
            return False

    def next(self):
        if self.ops_perf == "delete":
            if self.itr < self.num_deleted_keys:
                self._check_key_revId(self.deleted_keys[self.itr])
            self.itr += 1
        elif self.ops_perf == "update":
            if self.itr < self.num_valid_keys:
                self._check_key_revId(self.valid_keys[self.itr])
            self.itr += 1

    def _check_key_revId(self, key):
        try:
            src = self.client.memcached(key)
            dest = self.client_dest.memcached(key)
            seqno_src, cas_src, exp_src, flags_src = src.getRev(key)
            seqno_dest, cas_dest, exp_dest, flags_dest = dest.getRev(key)

            if seqno_src != seqno_dest:
                self.err_count += 1

                self.log.error(
                    "Mismatch on sequence numbers for key {0}\t Source Sequence Num:{1}\t Destination Sequence Num:{2}\tError Count{3}".format(
                        key, seqno_src, seqno_dest, self.err_count))
                self.state = FINISHED
            elif cas_src != cas_dest:
                self.err_count += 1
                self.log.error(
                    "Mismatch on CAS for key {0}\t Source CAS:{1}\t Destination CAS:{2}\tError Count{3}".format(key,
                        cas_src,
                        cas_dest, self.err_count))
                self.state = FINISHED
            elif exp_src != exp_dest:
                self.err_count += 1
                self.log.error(
                    "Mismatch on Expiry Flags for key {0}\t Source Expiry Flags:{1}\t Destination Expiry Flags:{2}\tError Count{3}".format(
                        key,
                        exp_src, exp_dest, self.err_count))
                self.state = FINISHED
            elif flags_src != flags_dest:
                self.err_count += 1
                self.log.error(
                    "Mismatch on Flags for key {0}\t Source Flags:{1}\t Destination Flags:{2}\tError Count{3}".format(
                        key,
                        flags_src, flags_dest, self.err_count))
                self.state = FINISHED
        except MemcachedError as error:
            if error.status == ERR_NOT_FOUND:
                pass
            else:
                self.state = FINISHED
                self.set_exception(error)

class ViewCreateTask(Task):
    def __init__(self, server, design_doc_name, view, bucket="default"):
        Task.__init__(self, "create_view_task")
        self.server = server
        self.bucket = bucket
        self.view = view
        prefix = ""
        if self.view:
            prefix = ("", "dev_")[self.view.dev_view]
        self.design_doc_name = prefix + design_doc_name
        self.ddoc_rev_no = 0

    def execute(self, task_manager):

        rest = RestConnection(self.server)

        try:
            # appending view to existing design doc
            content, meta = rest.get_ddoc(self.bucket, self.design_doc_name)
            ddoc = DesignDocument._init_from_json(self.design_doc_name, content)
            #if view is to be updated
            if self.view:
                ddoc.add_view(self.view)
            self.ddoc_rev_no = self._parse_revision(meta['rev'])

        except ReadDocumentException:
            # creating first view in design doc
            if self.view:
                ddoc = DesignDocument(self.design_doc_name, [self.view])
            #create an empty design doc
            else:
                ddoc = DesignDocument(self.design_doc_name, [])

        try:
            rest.create_design_document(self.bucket, ddoc)
            self.state = CHECKING
            task_manager.schedule(self)

        except DesignDocCreationException as e:
            self.state = FINISHED
            self.set_exception(e)

        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    def check(self, task_manager):
        rest = RestConnection(self.server)

        try:
            #only query if the DDoc has a view
            if self.view:
                query = {"stale" : "ok"}
                content = \
                    rest.query_view(self.design_doc_name, self.view.name, self.bucket, query)
                self.log.info("view : {0} was created successfully in ddoc: {1}".format(self.view.name, self.design_doc_name))
            else:
                #if we have reached here, it means design doc was successfully updated
                self.log.info("Design Document : {0} was updated successfully".format(self.design_doc_name))

            self.state = FINISHED
            if self._check_ddoc_revision():
                self.set_result(self.ddoc_rev_no)
            else:
                self.set_exception(Exception("failed to update design document"))
        except QueryViewException as e:
            task_manager.schedule(self, 2)

        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    def _check_ddoc_revision(self):
        valid = False
        rest = RestConnection(self.server)

        try:
            content, meta = rest.get_ddoc(self.bucket, self.design_doc_name)
            new_rev_id = self._parse_revision(meta['rev'])
            if new_rev_id != self.ddoc_rev_no:
                self.ddoc_rev_no = new_rev_id
                valid = True
        except ReadDocumentException:
            pass

        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

        return valid

    def _parse_revision(self, rev_string):
        return int(rev_string.split('-')[0])

class ViewDeleteTask(Task):
    def __init__(self, server, design_doc_name, view, bucket="default"):
        Task.__init__(self, "delete_view_task")
        self.server = server
        self.bucket = bucket
        self.view = view
        prefix = ""
        if self.view:
            prefix = ("", "dev_")[self.view.dev_view]
        self.design_doc_name = prefix + design_doc_name

    def execute(self, task_manager):

        rest = RestConnection(self.server)

        try:
            if self.view:
                # remove view from existing design doc
                content, header = rest.get_ddoc(self.bucket, self.design_doc_name)
                ddoc = DesignDocument._init_from_json(self.design_doc_name, content)
                status = ddoc.delete_view(self.view)
                if not status:
                    self.state = FINISHED
                    self.set_exception(Exception('View does not exist! %s' % (self.view.name)))

                # update design doc
                rest.create_design_document(self.bucket, ddoc)
                self.state = CHECKING
                task_manager.schedule(self, 2)
            else:
                #delete the design doc
                rest.delete_view(self.bucket, self.design_doc_name)
                self.log.info("Design Doc : {0} was successfully deleted".format(self.design_doc_name))
                self.state = FINISHED
                self.set_result(True)

        except (ValueError, ReadDocumentException, DesignDocCreationException) as e:
            self.state = FINISHED
            self.set_exception(e)

        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    def check(self, task_manager):
        rest = RestConnection(self.server)

        try:
            # make sure view was deleted
            query = {"stale" : "ok"}
            content = \
                rest.query_view(self.design_doc_name, self.view.name, self.bucket, query)
            self.state = FINISHED
            self.set_result(False)
        except QueryViewException as e:
            self.log.info("view : {0} was successfully deleted in ddoc: {1}".format(self.view.name, self.design_doc_name))
            self.state = FINISHED
            self.set_result(True)

        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

class ViewQueryTask(Task):
    def __init__(self, server, design_doc_name, view_name,
                 query, expected_rows=None,
                 bucket="default", retry_time=2):
        Task.__init__(self, "query_view_task")
        self.server = server
        self.bucket = bucket
        self.view_name = view_name
        self.design_doc_name = design_doc_name
        self.query = query
        self.expected_rows = expected_rows
        self.retry_time = retry_time

    def execute(self, task_manager):

        rest = RestConnection(self.server)

        try:
            # make sure view can be queried
            content = \
                rest.query_view(self.design_doc_name, self.view_name, self.bucket, self.query)

            if self.expected_rows is None:
                # no verification
                self.state = FINISHED
                self.set_result(content)
            else:
                self.state = CHECKING
                task_manager.schedule(self)

        except QueryViewException as e:
            # initial query failed, try again
            task_manager.schedule(self, self.retry_time)

        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    def check(self, task_manager):
        rest = RestConnection(self.server)

        try:
            # query and verify expected num of rows returned
            content = \
                rest.query_view(self.design_doc_name, self.view_name, self.bucket, self.query)

            self.log.info("(%d rows) expected, (%d rows) returned" % \
                          (self.expected_rows, len(content['rows'])))

            if len(content['rows']) == self.expected_rows:
                self.log.info("expected number of rows: '{0}' was found for view query".format(self.
                            expected_rows))
                self.state = FINISHED
                self.set_result(True)
            else:
                if "stale" in self.query:
                    if self.query["stale"].lower() == "false":
                        self.state = FINISHED
                        self.set_result(False)

                # retry until expected results or task times out
                task_manager.schedule(self, self.retry_time)
        except QueryViewException as e:
            # subsequent query failed! exit
            self.state = FINISHED
            self.set_exception(e)

        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)


class ModifyFragmentationConfigTask(Task):

    """
        Given a config dictionary attempt to configure fragmentation settings.
        This task will override the default settings that are provided for
        a given <bucket>.
    """

    def __init__(self, server, config=None, bucket="default"):
        Task.__init__(self, "modify_frag_config_task")

        self.server = server
        self.config = {"parallelDBAndVC" : "false",
                       "dbFragmentThreshold" : None,
                       "viewFragmntThreshold" : None,
                       "dbFragmentThresholdPercentage" : 100,
                       "viewFragmntThresholdPercentage" : 100,
                       "allowedTimePeriodFromHour" : None,
                       "allowedTimePeriodFromMin" : None,
                       "allowedTimePeriodToHour" : None,
                       "allowedTimePeriodToMin" : None,
                       "allowedTimePeriodAbort" : None,
                       "autoCompactionDefined" : "true"}
        self.bucket = bucket

        for key in config:
            self.config[key] = config[key]

    def execute(self, task_manager):
        rest = RestConnection(self.server)

        try:
            rest.set_auto_compaction(parallelDBAndVC=self.config["parallelDBAndVC"],
                                     dbFragmentThreshold=self.config["dbFragmentThreshold"],
                                     viewFragmntThreshold=self.config["viewFragmntThreshold"],
                                     dbFragmentThresholdPercentage=self.config["dbFragmentThresholdPercentage"],
                                     viewFragmntThresholdPercentage=self.config["viewFragmntThresholdPercentage"],
                                     allowedTimePeriodFromHour=self.config["allowedTimePeriodFromHour"],
                                     allowedTimePeriodFromMin=self.config["allowedTimePeriodFromMin"],
                                     allowedTimePeriodToHour=self.config["allowedTimePeriodToHour"],
                                     allowedTimePeriodToMin=self.config["allowedTimePeriodToMin"],
                                     allowedTimePeriodAbort=self.config["allowedTimePeriodAbort"],
                                     bucket=self.bucket)

            self.state = CHECKING
            task_manager.schedule(self, 10)

        except Exception as e:
            self.state = FINISHED
            self.set_exception(e)


    def check(self, task_manager):

        rest = RestConnection(self.server)
        try:
            # verify server accepted settings
            content = rest.get_bucket_json(self.bucket)
            if content["autoCompactionSettings"] == False:
                self.set_exception(Exception("Failed to set auto compaction settings"))
            else:
                # retrieved compaction settings
                self.set_result(True)
            self.state = FINISHED

        except GetBucketInfoFailed as e:
            # subsequent query failed! exit
            self.state = FINISHED
            self.set_exception(e)
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)


class MonitorViewFragmentationTask(Task):

    """
        Attempt to monitor fragmentation that is occurring for a given design_doc.
        execute stage is just for preliminary sanity checking of values and environment.

        Check function looks at index file accross all nodes and attempts to calculate
        total fragmentation occurring by the views within the design_doc.

        Note: If autocompaction is enabled and user attempts to monitor for fragmentation
        value higher than level at which auto_compaction kicks in a warning is sent and
        it is best user to use lower value as this can lead to infinite monitoring.
    """

    def __init__(self, server, design_doc_name, fragmentation_value=10, bucket="default"):

        Task.__init__(self, "monitor_frag_task")
        self.server = server
        self.bucket = bucket
        self.fragmentation_value = fragmentation_value
        self.design_doc_name = design_doc_name


    def execute(self, task_manager):

        # sanity check of fragmentation value
        if  self.fragmentation_value < 0 or self.fragmentation_value > 100:
            err_msg = \
                "Invalid value for fragmentation %d" % self.fragmentation_value
            self.state = FINISHED
            self.set_exception(Exception(err_msg))

        # warning if autocompaction is less than <fragmentation_value>
        try:
            auto_compact_percentage = self._get_current_auto_compaction_percentage()
            if auto_compact_percentage != "undefined" and auto_compact_percentage < self.fragmentation_value:
                self.log.warn("Auto compaction is set to %s. Therefore fragmentation_value %s may not be reached" % (auto_compact_percentage, self.fragmentation_value))

            self.state = CHECKING
            task_manager.schedule(self, 5)
        except GetBucketInfoFailed as e:
            self.state = FINISHED
            self.set_exception(e)
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)


    def _get_current_auto_compaction_percentage(self):
        """ check at bucket level and cluster level for compaction percentage """

        auto_compact_percentage = None
        rest = RestConnection(self.server)

        content = rest.get_bucket_json(self.bucket)
        if content["autoCompactionSettings"] != False:
            auto_compact_percentage = \
                content["autoCompactionSettings"]["viewFragmentationThreshold"]["percentage"]
        else:
            # try to read cluster level compaction settings
            content = rest.cluster_status()
            auto_compact_percentage = \
                content["autoCompactionSettings"]["viewFragmentationThreshold"]["percentage"]

        return auto_compact_percentage

    def check(self, task_manager):

        rest = RestConnection(self.server)
        new_frag_value = MonitorViewFragmentationTask.\
            calc_ddoc_fragmentation(rest, self.design_doc_name, self.bucket)

        self.log.info("current amount of fragmentation = %d" % new_frag_value)
        if new_frag_value > self.fragmentation_value:
            self.state = FINISHED
            self.set_result(True)
        else:
            # try again
            task_manager.schedule(self, 2)

    @staticmethod
    def aggregate_ddoc_info(rest, design_doc_name, bucket="default"):

        nodes = rest.node_statuses()
        info = []
        for node in nodes:
            server_info = {"ip" : node.ip,
                           "port" : node.port,
                           "username" : rest.username,
                           "password" : rest.password}
            rest = RestConnection(server_info)
            status, content = rest.set_view_info(bucket, design_doc_name)
            if status:
                info.append(content)

        return info
    @staticmethod
    def calc_ddoc_fragmentation(rest, design_doc_name, bucket="default"):

        total_disk_size = 0
        total_data_size = 0
        total_fragmentation = 0

        nodes_ddoc_info = \
            MonitorViewFragmentationTask.aggregate_ddoc_info(rest,
                                                         design_doc_name,
                                                         bucket)
        total_disk_size = sum([content['disk_size'] for content in nodes_ddoc_info])
        total_data_size = sum([content['data_size'] for content in nodes_ddoc_info])

        if total_disk_size > 0 and total_data_size > 0:
            total_fragmentation = \
                (total_disk_size - total_data_size) / float(total_disk_size) * 100

        return total_fragmentation

class ViewCompactionTask(Task):

    """
        Executes view compaction for a given design doc. This is technicially view compaction
        as represented by the api and also because the fragmentation is generated by the
        keys emitted by map/reduce functions within views.  Task will check that compaction
        history for design doc is incremented and if any work was really done.
    """

    def __init__(self, server, design_doc_name, bucket="default"):

        Task.__init__(self, "view_compaction_task")
        self.server = server
        self.bucket = bucket
        self.design_doc_name = design_doc_name
        self.ddoc_id = "_design%2f" + design_doc_name
        self.num_of_compactions = 0
        self.precompacted_frag_val = 0

    def execute(self, task_manager):
        rest = RestConnection(self.server)

        try:
            self.num_of_compactions, self.precompacted_frag_val = \
                self._get_compaction_details()
            rest.ddoc_compaction(self.ddoc_id)
            self.state = CHECKING
            task_manager.schedule(self, 2)
        except (CompactViewFailed, SetViewInfoNotFound) as ex:
            self.state = FINISHED
            self.set_exception(ex)
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    # verify compaction history incremented and some defraging occurred
    def check(self, task_manager):

        try:
            new_compaction_count, compacted_frag_val = self._get_compaction_details()
            if new_compaction_count == self.num_of_compactions and self._is_compacting():
                # compaction ran sucessfully but compactions was not changed
                # perhaps we are still compacting
                self.log.info("design doc {0} is compacting".format(self.design_doc_name))
                task_manager.schedule(self, 2)
            elif new_compaction_count > self.num_of_compactions:
                frag_val_diff = compacted_frag_val - self.precompacted_frag_val
                self.log.info("fragmentation went from %d to %d" % \
                              (self.precompacted_frag_val, compacted_frag_val))

                if frag_val_diff > 0:
                    # compaction ran sucessfully but datasize still same
                    # perhaps we are still compacting
                    if self._is_compacting():
                        task_manager.schedule(self, 2)

                    # probably we already compacted, but no work needed to be done
                    # returning False
                    self.set_result(False)
                else:
                    self.set_result(True)
                    self.state = FINISHED
            else:
                #Sometimes the compacting is not started immediately
                time.sleep(5)
                if self._is_compacting():
                    task_manager.schedule(self, 2)
                else:
                    self.set_exception(Exception("Check system logs, looks like compaction failed to start"))

        except (SetViewInfoNotFound) as ex:
            self.state = FINISHED
            self.set_exception(ex)
        #catch and set all unexpected exceptions
        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    def _get_compaction_details(self):
        rest = RestConnection(self.server)
        status, content = rest.set_view_info(self.bucket, self.design_doc_name)
        curr_no_of_compactions = content["stats"]["compactions"]
        curr_ddoc_fragemtation = \
            MonitorViewFragmentationTask.calc_ddoc_fragmentation(rest, self.design_doc_name, self.bucket)
        return (curr_no_of_compactions, curr_ddoc_fragemtation)

    def _is_compacting(self):
        rest = RestConnection(self.server)
        status, content = rest.set_view_info(self.bucket, self.design_doc_name)
        return content["compact_running"] == True

'''task class for failover. This task will only failover nodes but doesn't
 rebalance as there is already a task to do that'''
class FailoverTask(Task):
    def __init__(self, servers, to_failover=[], wait_for_pending=20):
        Task.__init__(self, "failover_task")
        self.servers = servers
        self.to_failover = to_failover
        self.wait_for_pending = wait_for_pending

    def execute(self, task_manager):
        try:
            self._failover_nodes(task_manager)
            self.log.info("{0} seconds sleep after failover, for nodes to go pending....".format(self.wait_for_pending))
            time.sleep(self.wait_for_pending)
            self.state = FINISHED
            self.set_result(True)

        except FailoverFailedException as e:
            self.state = FINISHED
            self.set_exception(e)

        except Exception as e:
            self.state = FINISHED
            self.log.info("Unexpected Exception Caught")
            self.set_exception(e)

    def _failover_nodes(self, task_manager):
        rest = RestConnection(self.servers[0])
        #call REST fail_over for the nodes to be failed over
        for server in self.to_failover:
            for node in rest.node_statuses():
                if server.ip == node.ip and int(server.port) == int(node.port):
                    self.log.info("Failing over {0}:{1}".format(node.ip, node.port))
                    rest.fail_over(node.id)

class GenerateExpectedViewResultsTask(Task):

    """
        Task to produce the set of keys that are expected to be returned
        by querying the provided <view>.  Results can be later passed to
        ViewQueryVerificationTask and compared with actual results from
        server.

        Currently only views with map functions that emit a single string
        or integer as keys are accepted.

        Also NOTE, this task is to be used with doc_generators that
        produce json like documentgenerator.DocumentGenerator
    """
    def __init__(self, doc_generators, view, query):
        Task.__init__(self, "generate_view_query_results_task")
        self.doc_generators = doc_generators
        self.view = view
        self.query = query
        self.emitted_rows = []


    def execute(self, task_manager):
        self.generate_emitted_rows()
        self.filter_emitted_rows()
        self.log.info("Finished generating expected query results")
        self.state = CHECKING
        task_manager.schedule(self)

    def check(self, task_manager):
        self.state = FINISHED
        self.set_result(self.emitted_rows)


    def generate_emitted_rows(self):
        for doc_gen in self.doc_generators:

            query_doc_gen = copy.deepcopy(doc_gen)
            while query_doc_gen.has_next():
                emit_key = re.sub(r',.*', '', re.sub(r'.*emit\([ +]?doc\.', '', self.view.map_func))
                _id, val = query_doc_gen.next()
                val = json.loads(val)

                val_emit_key = val[emit_key]
                if isinstance(val_emit_key, unicode):
                    val_emit_key = val_emit_key.encode('utf-8')
                self.emitted_rows.append({'id' : _id, 'key' : val_emit_key})

    def filter_emitted_rows(self):

        query = self.query

        # parse query flags
        descending_set = 'descending' in query and query['descending'] == "true"
        startkey_set, endkey_set = 'startkey' in query, 'endkey' in query
        startkey_docid_set, endkey_docid_set = 'startkey_docid' in query, 'endkey_docid' in query
        inclusive_end_false = 'inclusive_end' in query and query['inclusive_end'] == "false"
        key_set = 'key' in query

        # sort expected results to match view results
        expected_rows = sorted(self.emitted_rows,
                               cmp=GenerateExpectedViewResultsTask.cmp_result_rows,
                               reverse=descending_set)

        # filter rows according to query flags
        if startkey_set:
            start_key = query['startkey']
        else:
            start_key = expected_rows[0]['key']
        if endkey_set:
            end_key = query['endkey']
        else:
            end_key = expected_rows[-1]['key']

        if descending_set:
            start_key, end_key = end_key, start_key

        if startkey_set or endkey_set:
            if isinstance(start_key, str):
                start_key = start_key.strip("\"")
            if isinstance(end_key, str):
                end_key = end_key.strip("\"")
            expected_rows = [row for row in expected_rows if row['key'] >= start_key and row['key'] <= end_key]

        if key_set:
            key_ = query['key']
            start_key, end_key = key_, key_
            expected_rows = [row for row in expected_rows if row['key'] == key_]


        if descending_set:
            startkey_docid_set, endkey_docid_set = endkey_docid_set, startkey_docid_set

        if startkey_docid_set:
            if not startkey_set:
                self.log.warn("Ignoring startkey_docid filter when startkey is not set")
            else:
                do_filter = False
                if descending_set:
                    if endkey_docid_set:
                        startkey_docid = query['endkey_docid']
                        do_filter = True
                else:
                    startkey_docid = query['startkey_docid']
                    do_filter = True

                if do_filter:
                    expected_rows = \
                        [row for row in expected_rows if row['id'] >= startkey_docid or row['key'] > start_key]

        if endkey_docid_set:
            if not endkey_set:
                self.log.warn("Ignoring endkey_docid filter when endkey is not set")
            else:
                do_filter = False
                if descending_set:
                    if endkey_docid_set:
                        endkey_docid = query['startkey_docid']
                        do_filter = True
                else:
                    endkey_docid = query['endkey_docid']
                    do_filter = True

                if do_filter:
                    expected_rows = \
                        [row for row in expected_rows if row['id'] <= endkey_docid or row['key'] < end_key]


        if inclusive_end_false:
            if endkey_set and not endkey_docid_set:
                # remove all keys that match endkey
                expected_rows = [row for row in expected_rows['key'] if row['key'] != end_key]
            else:
                # remove last key
                expected_rows = expected_rows[:-1]

        self.emitted_rows = expected_rows

    @staticmethod
    def cmp_result_rows(x, y):
        rc = cmp(x['key'], y['key'])
        if rc == 0:
            # sort by id is tie breaker
            rc = cmp(x['id'], y['id'])
        return rc

class ViewQueryVerificationTask(Task):

    """
        * query with stale=false
        * check for duplicates
        * check for missing docs
            * check memcached
            * check couch
    """
    def __init__(self, server, design_doc_name, view_name, query, expected_rows,
                num_verified_docs=20, bucket="default", query_timeout=120):

        Task.__init__(self, "view_query_verification_task")
        self.server = server
        self.design_doc_name = design_doc_name
        self.view_name = view_name
        self.query = query
        self.expected_rows = expected_rows
        self.num_verified_docs = num_verified_docs
        self.bucket = bucket
        self.query_timeout = query_timeout
        self.results = None

        for key in config:
            self.config[key] = config[key]

    def execute(self, task_manager):

        rest = RestConnection(self.server)

        try:
            # query for full view results
            self.query["stale"] = "false"
            self.query["reduce"] = "false"
            self.query["include_docs"] = "true"
            self.results = rest.query_view(self.design_doc_name, self.view_name,
                                           self.bucket, self.query, timeout=self.query_timeout)
        except QueryViewException as e:
            self.set_exception(e)
            self.state = FINISHED


        msg = "Checking view query results: (%d keys expected) vs (%d keys returned)" % \
            (len(self.expected_rows), len(self.results['rows']))
        self.log.info(msg)

        self.state = CHECKING
        task_manager.schedule(self)

    def check(self, task_manager):
        err_infos = []
        rc_status = {"passed" : False,
                     "errors" : err_infos}  # array of dicts with keys 'msg' and 'details'

        # create verification id lists
        expected_ids = [row['id'] for row in self.expected_rows]
        couch_ids = [str(row['id']) for row in self.results['rows']]

        # check results
        self.check_for_duplicate_ids(expected_ids, couch_ids, err_infos)
        self.check_for_missing_ids(expected_ids, couch_ids, err_infos)
        self.check_for_value_corruption(err_infos)

        # check for errors
        if len(rc_status["errors"]) == 0:
           rc_status["passed"] = True

        self.state = FINISHED
        self.set_result(rc_status)

    def check_for_duplicate_ids(self, expected_ids, couch_ids, err_infos):

        extra_id_set = set(couch_ids) - set(expected_ids)

        seen = set()
        for id in couch_ids:
            if id in seen and id not in extra_id_set:
                extra_id_set.add(id)
            else:
                seen.add(id)

        if len(extra_id_set) > 0:
            # extra/duplicate id verification
            dupe_rows = [row for row in self.results['rows'] if row['id'] in extra_id_set]
            err = { "msg" : "duplicate rows found in query results",
                    "details" : dupe_rows }
            err_infos.append(err)

    def check_for_missing_ids(self, expected_ids, couch_ids, err_infos):

        missing_id_set = set(expected_ids) - set(couch_ids)

        if len(missing_id_set) > 0:

            missing_id_errors = self.debug_missing_items(missing_id_set)

            if len(missing_id_errors) > 0:
                err = { "msg" : "missing ids from memcached",
                        "details" : missing_id_errors}
                err_infos.append(err)


    def check_for_value_corruption(self, err_infos):

        if self.num_verified_docs > 0:

            doc_integrity_errors = self.include_doc_integrity()

            if len(doc_integrity_errors) > 0:
                err = { "msg" : "missmatch in document values",
                        "details" : doc_integrity_errors }
                err_infos.append(err)


    def debug_missing_items(self, missing_id_set):
        rest = RestConnection(self.server)
        client = KVStoreAwareSmartClient(rest, self.bucket)
        missing_id_errors = []

        # debug missing documents
        for doc_id in list(missing_id_set)[:self.num_verified_docs]:

            # attempt to retrieve doc from memcached
            mc_item = client.mc_get_full(doc_id)
            if mc_item == None:
                missing_id_errors.append("document %s missing from memcached" % (doc_id))

            # attempt to retrieve doc from disk
            else:

                num_vbuckets = len(rest.get_vbuckets(self.bucket))
                doc_meta = client.get_doc_metadata(num_vbuckets, doc_id)

                if(doc_meta != None):
                    if (doc_meta['key_valid'] != 'valid'):
                        msg = "Error expected in results for key with invalid state %s" % doc_meta
                        missing_id_errors.append(msg)

                else:
                    msg = "query doc_id: %s doesn't exist in bucket: %s" % \
                        (doc_id, self.bucket)
                    missing_id_errors.append(msg)

            if(len(missing_id_errors) == 0):
                msg = "view engine failed to index doc [%s] in query: %s" % (doc_id, self.query)
                missing_id_errors.append(msg)

            return missing_id_errors

    def include_doc_integrity(self):
        rest = RestConnection(self.server)
        client = KVStoreAwareSmartClient(rest, self.bucket)
        doc_integrity_errors = []

        exp_verify_set = [row['doc'] for row in\
            self.results['rows'][:self.num_verified_docs]]

        for view_doc in exp_verify_set:
            doc_id = str(view_doc['_id'])
            mc_item = client.mc_get_full(doc_id)

            if mc_item is not None:
                mc_doc = json.loads(mc_item["value"])

                # compare doc content
                for key in mc_doc.keys():
                    if(mc_doc[key] != view_doc[key]):
                        err_msg = \
                            "error verifying document id %s: retrieved value %s expected %s \n" % \
                                (doc_id, mc_doc[key], view_doc[key])
                        doc_integrity_errors.append(err_msg)
            else:
                doc_integrity_errors.append("doc_id %s could not be retrieved for verification \n" % doc_id)

        return doc_integrity_errors



#!/usr/bin/env python

import re
import sys
import math
import time
import socket
import string
import struct
import random
import threading
import multiprocessing

sys.path.append("lib")
sys.path.append(".")

try:
    import logging
    logging.config.fileConfig("mcsoda.logging.conf")
    log = logging.getLogger()
except:
    class P:
        def error(self, m): print(m)
        def info(self, m):  print(m)
    log = P()

try:
    from hashlib import md5
    md5; # Pyflakes workaround
except ImportError:
    from md5 import md5

import crc32
import mc_bin_client

from membase.api.exception import QueryViewException

from memcacheConstants import REQ_MAGIC_BYTE, RES_MAGIC_BYTE
from memcacheConstants import ERR_NOT_MY_VBUCKET, ERR_ENOMEM, ERR_EBUSY, ERR_ETMPFAIL
from memcacheConstants import REQ_PKT_FMT, RES_PKT_FMT, MIN_RECV_PACKET
from memcacheConstants import SET_PKT_FMT, CMD_GET, CMD_SET, CMD_DELETE
from memcacheConstants import CMD_ADD, CMD_REPLACE, CMD_PREPEND, CMD_APPEND # "ARPA"

from libobserve.obs_mcsoda import McsodaObserver
from libobserve.obs import Observable
from libobserve.obs_helper import UnblockingJoinableQueue
from libstats.carbon_feeder import CarbonFeeder
from libstats.carbon_key import CarbonKey

LARGE_PRIME = 9576890767

# --------------------------------------------------------

INT_TYPE = type(123)
FLOAT_TYPE = type(0.1)
DICT_TYPE = type({})

def dict_to_s(d, level="", res=None, suffix=", ", ljust=None):
    res = res or []
    return ''.join(dict_to_s_inner(d, level, res, suffix, ljust))

def dict_to_s_inner(d, level, res, suffix, ljust):
    dtype = DICT_TYPE
    scalars = []
    complex = []

    for key in d.keys():
        if type(d[key]) == dtype:
            complex.append(key)
        else:
            scalars.append(key)
    scalars.sort()
    complex.sort()

    # Special case for histogram output.
    histo_max = 0
    histo_sum = 0
    if scalars and not complex and \
        type(scalars[0]) == FLOAT_TYPE and type(d[scalars[0]]) == INT_TYPE:
        for key in scalars:
            v = d[key]
            histo_max = max(v, histo_max)
            histo_sum = histo_sum + v

    histo_cur = 0 # Running total for histogram output.
    for key in scalars:
        if type(key) == FLOAT_TYPE:
            k = re.sub("0*$", "", "%.7f" % (key))
        else:
            k = str(key)
        if ljust:
            k = string.ljust(k, ljust)
        x = d[key]
        if histo_max:
            histo_cur = histo_cur + x
        v = str(x)
        if histo_max:
            v = string.rjust(v, 8) + " " + \
                string.rjust("{0:.1%}".format(histo_cur / float(histo_sum)), 8) + " " + \
                ("*" * int(math.ceil(50.0 * d[key] / histo_max)))

        res.append(level + k + ": " + v + suffix)

    # Recurse for nested, dictionary values.
    if complex:
        res.append("\n")
    for key in complex:
        res.append(level   + str(key) + ":\n")
        dict_to_s_inner(d[key], level + "  ", res, "\n", 9)

    return res

# The histo dict is returned by add_timing_sample().
# The percentiles must be sorted, ascending, like [0.90, 0.99].
def histo_percentile(histo, percentiles):
    v_sum = 0
    bins = histo.keys()
    bins.sort()
    for bin in bins:
        v_sum += histo[bin]
    v_sum = float(v_sum)
    v_cur = 0 # Running total.
    rv = []
    for bin in bins:
        if not percentiles:
            return rv
        v_cur += histo[bin]
        while percentiles and (v_cur / v_sum) >= percentiles[0]:
            rv.append((percentiles[0], bin))
            percentiles.pop(0)
    return rv

# --------------------------------------------------------

MIN_VALUE_SIZE = [10]

def obs_cb(store):
    """
    callback for observe thread.
    """
    if not store:
        print "[mcsoda] obs_cb is broken"
        return

    print "[mcsoda] obs_cb: clear obs_key_cas %s" % store.obs_key_cas
    store.obs_key_cas.clear()

def woq_worker(req_queue, stats_queue, ctl, cfg, store):
    """
    measure latencies of standard write/observe/query patterns
    """
    bucket = "default"
    ddoc = "A"
    view = "city1"  # TODO pass from eperf
    query_params = {"limit": 10,
                    "stale": "false"}

    print "[mcsoda] woq_worker started"
    woq_observer = McsodaObserver(ctl, cfg, store, None)

    while True:

        key, cas = req_queue.get(block=True)
        start_time = time.time() # latency includes observe and query time

        # observe
        if not woq_observer.block_for_persistence(key, cas):
            # put an invalid object to indicate error
            stats_queue.put([key, cas, 0, 0, 0, 0], block=True)
            req_queue.task_done()
            continue

        obs_latency = time.time() - start_time
        if cfg.get("woq-verbose", 0):
            print "[mcsoda] woq_worker obs latency: %s, key = %s, cas = %s "\
                % (obs_latency, key, cas)

        query_start = time.time()

        try:
            result = store.rest.query_view(ddoc, view, bucket, query_params)
        except QueryViewException as e:
            print "[mcsoda] woq_worker QueryViewException: %s" % e
            stats_queue.put([key, cas, 0, 0, 0, 0], block=True)
            req_queue.task_done()
            continue

        query_latency = time.time() - query_start
        if cfg.get("woq-verbose", 0):
            print "[mcsoda] woq_worker query latency: %s, key = %s, cas = %s "\
                % (query_latency, key, cas)
            print "[mcsoda] woq_worker query result: %s" % result

        latency = time.time() - start_time
        stats_queue.put([key, cas, start_time, obs_latency, query_latency, latency],
                        block=True)
        req_queue.task_done()
    print "[mcsoda] woq_worker stopped working"

def run_worker(ctl, cfg, cur, store, prefix, heartbeat = 0, why = ""):
    i = 0
    t_last_flush = time.time()
    t_last_cycle = time.time()
    o_last_flush = store.num_ops(cur)
    t_last = time.time()
    o_last = store.num_ops(cur)
    xfer_sent_last = 0
    xfer_recv_last = 0
    store.why = why
    store.stats_ops = cfg.get("stats_ops", 10000)
    if cfg.get('carbon', 0):
        store.c_feeder = \
            CarbonFeeder(cfg.get('carbon-server', '127.0.0.1'),
                         port=cfg.get('carbon-port', 2003),
                         timeout=cfg.get('carbon-timeout', 5),
                         cache_size=cfg.get('carbon-cache-size', 10))

    report = cfg.get('report', 0)
    hot_shift = cfg.get('hot-shift', 0)
    max_ops_per_sec = float(cfg.get('max-ops-per-sec', 0))

    if cfg.get('max-ops-per-sec', 0) > 0 and not 'batch' in cur:
        cur['batch'] = 10

    log.info("[mcsoda: %s] starts cfg: %s" %(why, cfg))
    log.info("[mcsoda: %s] starts cur: %s" %(why, cur))
    log.info("[mcsoda: %s] starts store: %s" %(why, store))
    log.info("[mcsoda: %s] starts prefix: %s" %(why, prefix))
    log.info("[mcsoda: %s] starts running." %why)

    heartbeat_last = t_last

    if cfg.get('woq-pattern', 0):
        woq_req_queue = UnblockingJoinableQueue(1)    # pattern: write/observe/query
        woq_stats_queue = multiprocessing.Queue(1)
        woq_process = multiprocessing.Process(target=woq_worker,
                                              args=(woq_req_queue, woq_stats_queue,
                                                    ctl, cfg, store))
        woq_process.daemon = True
        woq_process.start()

    if cfg.get('observe', 0):
        observer = McsodaObserver(ctl, cfg, store, obs_cb)
        observer.start()

    while ctl.get('run_ok', True):
        num_ops = cur.get('cur-gets', 0) + cur.get('cur-sets', 0)

        if cfg.get('max-ops', 0) > 0 and cfg.get('max-ops', 0) <= num_ops:
            break
        if cfg.get('exit-after-creates', 0) > 0 and \
            cfg.get('max-creates', 0) > 0 and \
            cfg.get('max-creates', 0) <= cur.get('cur-creates', 0):
            break
        if ctl.get('shutdown_event') is not None:
            if ctl['shutdown_event'].is_set():
                break

        heartbeat_duration = time.time() - heartbeat_last
        if heartbeat != 0 and heartbeat_duration > heartbeat:
            heartbeat_last += heartbeat_duration
            log.info("[mcsoda: %s] num_ops = %s. duration = %s" %(why, num_ops, heartbeat_duration))

        command = next_cmd(cfg, cur, store)
        flushed = store.command(command)

        if flushed and cfg.get('woq-pattern', 0):

            # record stats
            if not woq_stats_queue.empty():
                try:
                    key, cas, start_time, obs_latency, query_latency, latency \
                        = woq_stats_queue.get(block=False)
                    if not start_time and not latency:
                        store.woq_key_cas.clear()   # error
                    else:
                        store.add_timing_sample("woq-obs", obs_latency)
                        store.add_timing_sample("woq-query", query_latency)
                        store.add_timing_sample("woq", latency)
                        store.save_stats(start_time)
                        store.woq_key_cas.clear()   # simply clear all, no key/cas sanity check
                        print "[mcsoda] woq_stats: key: %s, cas: %s, " \
                            "obs_latency: %f, query_latency: %f, latency: %f" \
                            % (key, cas, obs_latency, query_latency, latency)
                except Queue.Empty:
                    pass

            # produce request
            if woq_req_queue.all_finished():
                for key_num, cas in store.woq_key_cas.iteritems():
                    key = prepare_key(key_num, cfg.get('prefix', ''))
                    try:
                        woq_req_queue.put([key, cas], block=False)
                    except Queue.Full:
                        break

        if flushed and cfg.get('observe', 0):
            if store.obs_key_cas and not observer.num_observables():
                observables = []
                for key_num, cas in store.obs_key_cas.iteritems():
                    obs = Observable(key=prepare_key(key_num, cfg.get('prefix', '')),
                                     cas=cas,
                                     persist_count=cfg.get('obs-persist-count', 1),
                                     repl_count=cfg.get('obs-repl-count', 1))
                    observables.append(obs)
                observer.load_observables(observables)

        i += 1

        if report > 0 and i % report == 0:
            t_curr = time.time()
            o_curr = store.num_ops(cur)
            xfer_sent_curr = store.xfer_sent
            xfer_recv_curr = store.xfer_recv

            t_delta = t_curr - t_last
            o_delta = o_curr - o_last
            xfer_sent_delta = xfer_sent_curr - xfer_sent_last
            xfer_recv_delta = xfer_recv_curr - xfer_recv_last

            ops_per_sec = o_delta / t_delta
            xfer_sent_per_sec = xfer_sent_delta / t_delta
            xfer_recv_per_sec = xfer_recv_delta / t_delta

            log.info(prefix + dict_to_s(cur))
            log.info("%s  ops: %s secs: %s ops/sec: %s" \
                     " tx-bytes/sec: %s rx-bytes/sec: %s" %
                     (prefix,
                      string.ljust(str(o_delta), 10),
                      string.ljust(str(t_delta), 15),
                      string.ljust(str(int(ops_per_sec)), 10),
                      string.ljust(str(int(xfer_sent_per_sec) or "unknown"), 10),
                      int(xfer_recv_per_sec) or "unknown"))

            t_last = t_curr
            o_last = o_curr
            xfer_sent_last = xfer_sent_curr
            xfer_recv_last = xfer_recv_curr

        if flushed:
            """Code below is responsible for speed limitation.
            Stream looks like ^_^_^_^_^_^_^

            delta1 = flush time + previous sleep time (^_)
            delta2 = flush time (^)

            TODO: dynamic correction factor. We have to measure actual average
            throughput - let's say - every minute. Thus we can adjust request
            rate. For now it's empiric, because we always oversleep.
            """
            CORRECTION_FACTOR = 0.975

            delta1 = time.time() - t_last_cycle
            delta2 = time.time() - t_last_flush
            t_last_cycle += delta1

            ops_done = float(store.num_ops(cur) - o_last_flush)
            o_last_flush += ops_done

            if max_ops_per_sec:
                # Taking into account global throughtput
                if cfg.get('active_fg_workers') is not None:
                    concurrent_workers = cfg.get('active_fg_workers').value
                else:
                    concurrent_workers = 1
                local_max_ops_per_sec = max_ops_per_sec / concurrent_workers
                # Actual throughput
                ops_per_sec = ops_done / delta2
                # Sleep if too fast. It must be too fast.
                if ops_per_sec > local_max_ops_per_sec:
                    sleep_time = CORRECTION_FACTOR * ops_done / local_max_ops_per_sec - delta2
                    time.sleep(max(sleep_time, 0))

            if hot_shift > 0:
                cur['cur-base'] = cur.get('cur-base', 0) + (hot_shift * delta1)

            t_last_flush = time.time()

    store.flush()

def next_cmd(cfg, cur, store):
    itm_val = None
    num_ops = cur.get('cur-ops', 0)

    do_set = cfg.get('ratio-sets', 0) > float(cur.get('cur-sets', 0)) / positive(num_ops)
    if do_set:
        itm_gen = True

        cmd = 'set'
        cur_sets = cur.get('cur-sets', 0) + 1
        cur['cur-sets'] = cur_sets
        cur['cur-ops'] = cur.get('cur-ops', 0) + 1

        do_set_create = ((cfg.get('max-items', 0) <= 0 or
                          cfg.get('max-items', 0) > cur.get('cur-items', 0)) and
                         cfg.get('max-creates', 0) > cur.get('cur-creates', 0) and
                         cfg.get('ratio-creates', 0) >= \
                           float(cur.get('cur-creates', 0)) / positive(cur.get('cur-sets', 0)))
        if do_set_create:
            # Create...
            key_num = cur.get('cur-items', 0)

            cur['cur-items'] = cur.get('cur-items', 0) + 1
            cur['cur-creates'] = cur.get('cur-creates', 0) + 1
        else:
            # Update...
            num_updates = cur['cur-sets'] - cur.get('cur-creates', 0)

            do_delete = cfg.get('ratio-deletes', 0) > \
                          float(cur.get('cur-deletes', 0)) / positive(num_updates)
            if do_delete:
                itm_gen = False
                cmd = 'delete'
                cur['cur-deletes'] = cur.get('cur-deletes', 0) + 1
            else:
                num_mutates = num_updates - cur.get('cur-deletes', 0)

                do_arpa = cfg.get('ratio-arpas', 0) > \
                          float(cur.get('cur-arpas', 0)) / positive(num_mutates)
                if do_arpa:
                    cmd = 'arpa'
                    cur['cur-arpas'] = cur.get('cur-arpas', 0) + 1

            key_num = choose_key_num(num_updates,
                                     cfg.get('ratio-hot', 0),
                                     cfg.get('ratio-hot-sets', 0),
                                     cur.get('cur-sets', 0),
                                     cur.get('cur-base', 0),
                                     cfg.get('random', 0),
                                     cur)

        expiration = 0
        if cmd[0] == 's' and cfg.get('ratio-expirations', 0.0) * 100 > cur_sets % 100:
            expiration = cfg.get('expiration', 0)

        key_str = prepare_key(key_num, cfg.get('prefix', ''))
        if itm_gen:
            itm_val = store.gen_doc(key_num, key_str,
                                    choose_entry(cfg.get('min-value-size', MIN_VALUE_SIZE),
                                                 num_ops))

        return (cmd, key_num, key_str, itm_val, expiration)
    else:
        cmd = 'get'
        cur['cur-gets'] = cur.get('cur-gets', 0) + 1
        cur['cur-ops'] = cur.get('cur-ops', 0) + 1

        do_query = cfg.get('ratio-queries', 0) > \
            float(cur.get('cur-queries', 0)) / cur.get('cur-gets', 0)
        if do_query:
            cmd = 'query'
            cur['cur-queries'] = cur.get('cur-queries', 0) + 1

        do_get_hit = (cfg.get('ratio-misses', 0) * 100) <= (cur.get('cur-gets', 0) % 100)
        if do_get_hit:
            key_num = choose_key_num(cur.get('cur-items', 0),
                                     cfg.get('ratio-hot', 0),
                                     cfg.get('ratio-hot-gets', 0),
                                     cur.get('cur-gets', 0),
                                     cur.get('cur-base', 0),
                                     cfg.get('random', 0),
                                     cur)
            key_str = prepare_key(key_num, cfg.get('prefix', ''))

            return (cmd, key_num, key_str, itm_val, 0)
        else:
            cur['cur-misses'] = cur.get('cur-misses', 0) + 1
            return (cmd, -1, prepare_key(-1, cfg.get('prefix', '')), None, 0)

def choose_key_num(num_items, ratio_hot, ratio_hot_choice,
                   num_ops, base, random_key, cur):
    """
    Choose a random or deterministic number in order to generate the MD5 hash.

    The deterministic algorithm always favors new items.
    i.e:
        If many items have been created (num_creates > num_hot_items), \
        hot items are chosen from the newest guys.
    """
    num_creates = cur.get('cur-creates', 0)
    if num_items < 0 \
        or ratio_hot < 0 or ratio_hot > 1:
        print "[mcsoda choose_key_num error] num_items: {0}, num_creates:{1}, ratio_hot: {2}"\
            .format(num_items, num_creates, ratio_hot)
        return 1

    # get a random or deterministic key
    if random_key == 1:
        x = int(random.random() * num_items)
    else:
        pos = cur.get('pos', 0)
        pos = (pos + LARGE_PRIME) % positive(num_items)
        cur['pos'] = pos
        x = pos

    hit_hot_range = (ratio_hot_choice * 100) > (num_ops % 100)
    num_hot_items = positive(math.floor(ratio_hot * num_items))
    num_cold_items = positive(num_items - num_hot_items)
    num_init_items = positive(num_items - num_creates)
    base = base % num_init_items

    # calculate offset and apply it to the base
    if hit_hot_range:
        offset = x % num_hot_items
        if offset > num_creates:                          # choose from the left hot set
            retval = (base + offset - num_creates) % num_init_items
        else:
            retval = num_items - offset                   # choose from the right hot set
    else:
        offset = x % num_cold_items
        if num_creates > num_hot_items:
            retval = offset
        elif base > num_cold_items:                         # no split-up on the cold set
            retval = (base + num_hot_items - num_creates + offset) % num_init_items
        elif offset < base:                                 # choose from the left cold set
            retval = offset
        else:
            retval = offset + num_hot_items - num_creates   # choose from the right cold set

    return int(retval) % num_items

def positive(x):
    if x > 0:
        return x
    return 1

def prepare_key(key_num, prefix=None):
    key_hash = md5(str(key_num)).hexdigest()[0:16]
    if prefix and len(prefix) > 0:
        return prefix + "-" + key_hash
    return key_hash

def choose_entry(arr, n):
    return arr[n % len(arr)]

# --------------------------------------------------------


class Store(object):


    def connect(self, target, user, pswd, cfg, cur, bucket="default"):
        self.target = target
        self.cfg = cfg
        self.cur = cur
        self.xfer_sent = 0
        self.xfer_recv = 0

    def show_some_keys(self):
        log.info("first 5 keys...")
        for i in range(5):
            print("echo get %s | nc %s %s" %
                 (self.cmd_line_get(i, prepare_key(i, self.cfg.get('prefix', ''))),
                  self.target.split(':')[0],
                  self.target.split(':')[1]))

    def stats_collector(self, sc):
        self.sc = sc

    def command(self, c):
        cmd, key_num, key_str, data, expiration = c
        if cmd[0] == 'g' or cmd[0] == 'q':
            print(cmd + ' ' + key_str + '\r')
            return False
        if cmd[0] == 'd':
            print('delete ' + key_str + '\r')
            return False

        c = 'set'
        if cmd[0] == 'a':
            c = self.arpa[self.cur.get('cur-sets', 0) % len(self.arpa)]

        print("%s %s 0 %s %s\r\n%s\r" % (c, key_str, expiration,
                                         len(data), data))
        return False

    def flush(self):
        pass

    def num_ops(self, cur):
        return cur.get('cur-gets', 0) + cur.get('cur-sets', 0)

    def gen_doc(self, key_num, key_str, min_value_size, json=None, cache=None):
        if json is None:
            json = self.cfg.get('json', 1) > 0
        if cache is None:
            cache = self.cfg.get('doc-cache', 0)

        return gen_doc_string(key_num, key_str, min_value_size,
                              self.cfg['suffix'][min_value_size],
                              json, cache=cache)

    def cmd_line_get(self, key_num, key_str):
        return key_str

    def readbytes(self, skt, nbytes, buf):
        while len(buf) < nbytes:
            data = None
            try:
                data = skt.recv(max(nbytes - len(buf), 4096))
            except socket.timeout:
                log.error("[mcsoda] EXCEPTION: Store.readbytes-socket.timeout / recv timed out")
                self.cur["cur-ex-Store.readbytes-socket.timeout"] = self.cur.get("cur-ex-Store.readbytes-socket.timeout", 0) + 1
            except Exception as e:
                log.error("[mcsoda] EXCEPTION: Store.readbytes: " + str(e))
                self.cur["cur-ex-Store.readbytes"] = self.cur.get("cur-ex-Store.readbytes", 0) + 1
            if not data:
                log.error("[mcsoda] Store.readbytes-nodata / skt.read no data.")
                self.cur["cur-Store.readbytes-nodata"] = self.cur.get("cur-Store.readbytes-odata", 0) + 1
                return None, ''
            buf += data
        return buf[:nbytes], buf[nbytes:]

    def add_timing_sample(self, cmd, delta, prefix="latency-"):
        base = prefix + cmd
        for suffix in self.cfg.get("timing-suffixes", ["", "-recent"]):
            key = base + suffix
            histo = self.cur.get(key, None)
            if histo is None:
                histo = {}
                self.cur[key] = histo
            try:
                bucket = round(self.histo_bucket(delta), 6)
                histo[bucket] = histo.get(bucket, 0) + 1
            except TypeError as e:
                print "[mcsoda TypeError] {0}, delta = {1}".format(str(e), delta)

    def histo_bucket(self, samp):
        hp = self.cfg.get("histo-precision", 2)
        if samp > 0:
            p = 10 ** (math.floor(math.log10(samp)) - (hp - 1))
            r = round(samp / p)
            return r * p

    def drange(self, start, stop, step):
        r = start
        while r < stop:
            yield round(float(r), 6)
            r += float(step)


class StoreMemcachedBinary(Store):

    def connect(self, target, user, pswd, cfg, cur, bucket="default"):
        self.cfg = cfg
        self.cur = cur
        self.target = target
        self.host_port = (target + ":11211").split(':')[0:2]
        self.host_port[1] = int(self.host_port[1])
        self.connect_host_port(self.host_port[0], self.host_port[1], user, pswd, bucket=bucket)
        self.inflight_reinit()
        self.queue = []
        self.cmds = 0
        self.ops = 0
        self.previous_ops = 0
        self.buf = ''
        self.arpa = [ (CMD_ADD,     True),
                      (CMD_REPLACE, True),
                      (CMD_APPEND,  False),
                      (CMD_PREPEND, False) ]
        self.xfer_sent = 0
        self.xfer_recv = 0
        self.obs_key_cas = {} # {key_num: cas} pair
        self.woq_key_cas = {} # {key_num: cas} pair

    def connect_host_port(self, host, port, user, pswd, bucket="default"):
        self.conn = mc_bin_client.MemcachedClient(host, port)
        if user:
            self.conn.sasl_auth_plain(user, pswd)

    def inflight_reinit(self, inflight=0):
        self.inflight = inflight
        self.inflight_num_gets = 0
        self.inflight_num_sets = 0
        self.inflight_num_deletes = 0
        self.inflight_num_arpas = 0
        self.inflight_num_queries = 0
        self.inflight_start_time = 0
        self.inflight_end_time = 0
        self.inflight_grp = None

    def inflight_start(self):
        return []

    def inflight_complete(self, inflight_arr):
        return ''.join(inflight_arr)

    def inflight_send(self, inflight_msg):
        self.conn.s.send(inflight_msg)
        return len(inflight_msg)

    def inflight_recv(self, inflight, inflight_arr, expectBuffer=None):
        received = 0
        for i in range(inflight):
            cmd, keylen, extralen, errcode, datalen, opaque, val, buf = self.recvMsg()
            received += datalen + MIN_RECV_PACKET
        return received

    def inflight_append_buffer(self, grp, vbucketId, opcode, opaque):
        return grp

    def command(self, c):
        self.queue.append(c)
        if len(self.queue) > self.flush_level():
            self.flush()
            return True
        return False

    def flush_level(self):
        return self.cur.get('batch') or \
               self.cfg.get('batch', 100)

    def get_vbucketId(self, key):
        vbuckets = self.cfg.get("vbuckets", 0)
        if vbuckets > 0:
            return crc32.crc32_hash(key) & (vbuckets - 1)
        return 0

    def header(self, op, key, val, opaque=0, extra='', cas=0,
               dtype=0,
               fmt=REQ_PKT_FMT,
               magic=REQ_MAGIC_BYTE):
        vbucketId = self.get_vbucketId(key)
        return struct.pack(fmt, magic, op,
                           len(key), len(extra), dtype, vbucketId,
                           len(key) + len(extra) + len(val), opaque, cas), vbucketId

    def create_seed(self):
        """Return a seed (hashable tuple or int value) based on current stats.
        This seed ensures reproducible randomness for the same test
        configurations.

        """

        if self.why == 'loop-fg':
            return self.cur.get('cur-queries', 0)
        else:
            return (self.cur.get('cur-gets', 0),
                    self.cur.get('cur-sets', 0),
                    self.cur.get('cur-deletes', 0),
                    self.cur.get('cur-creates', 0),
                    self.cur.get('cur-arpas', 0))

    def flush(self):
        next_inflight = 0
        next_inflight_num_gets = 0
        next_inflight_num_sets = 0
        next_inflight_num_deletes = 0
        next_inflight_num_arpas = 0
        next_inflight_num_queries = 0

        next_grp = self.inflight_start()

        # Permutation of requests
        random.seed(self.create_seed())
        random.shuffle(self.queue)

        # Start a 1, not 0, due to the single latency measurement request.
        for i in range(1, len(self.queue)):
            cmd, key_num, key_str, data, expiration = self.queue[i]
            delta_gets, delta_sets, delta_deletes, delta_arpas, delta_queries = \
                self.cmd_append(cmd, key_num, key_str, data, expiration, next_grp)
            next_inflight += 1
            next_inflight_num_gets += delta_gets
            next_inflight_num_sets += delta_sets
            next_inflight_num_deletes += delta_deletes
            next_inflight_num_arpas += delta_arpas
            next_inflight_num_queries += delta_queries

        next_msg = self.inflight_complete(next_grp)

        latency_cmd = None
        latency_start = 0
        latency_end = 0

        delta_gets = 0
        delta_sets = 0
        delta_deletes = 0
        delta_arpas = 0
        delta_queries = 0

        if self.inflight > 0:
            # Receive replies from the previous batch of inflight requests.
            self.xfer_recv += self.inflight_recv(self.inflight, self.inflight_grp)
            self.inflight_end_time = time.time()
            self.ops += self.inflight
            if self.sc:
                self.sc.ops_stats({ 'tot-gets':    self.inflight_num_gets,
                                    'tot-sets':    self.inflight_num_sets,
                                    'tot-deletes': self.inflight_num_deletes,
                                    'tot-arpas':   self.inflight_num_arpas,
                                    'tot-queries': self.inflight_num_queries,
                                    'start-time':  self.inflight_start_time,
                                    'end-time':    self.inflight_end_time })

        if len(self.queue) > 0:
            # Use the first request in the batch to measure single
            # request latency.
            grp = self.inflight_start()
            latency_cmd, key_num, key_str, data, expiration = self.queue[0]
            delta_gets, delta_sets, delta_deletes, delta_arpas, delta_queries = \
                self.cmd_append(latency_cmd,
                                key_num, key_str, data, expiration, grp)
            msg = self.inflight_complete(grp)

            latency_start = time.time()
            self.xfer_sent += self.inflight_send(msg)
            self.xfer_recv += self.inflight_recv(1, grp, expectBuffer=False)
            latency_end = time.time()

            self.ops += 1

        self.queue = []

        self.inflight_reinit()
        if next_inflight > 0:
            self.inflight = next_inflight
            self.inflight_num_gets = next_inflight_num_gets
            self.inflight_num_sets = next_inflight_num_sets
            self.inflight_num_deletes = next_inflight_num_deletes
            self.inflight_num_arpas = next_inflight_num_arpas
            self.inflight_num_queries = next_inflight_num_queries
            self.inflight_start_time = time.time()
            self.inflight_grp = next_grp
            self.xfer_sent += self.inflight_send(next_msg)

        if latency_cmd:
            delta = latency_end - latency_start
            self.add_timing_sample(latency_cmd, delta)
            if self.cfg.get('carbon', 0):
                if self.__class__.__name__ == "StoreMemcachedBinary":
                    server = self.conn.host
                else:
                    vbucketId = self.get_vbucketId(key_str)
                    server = self.awareness.vBucketMap[vbucketId]
                c_key = CarbonKey("mcsoda", server,
                                  "latency-" + latency_cmd)
                self.c_feeder.feed(c_key, delta * 1000)

        if self.sc:
            if self.ops - self.previous_ops > self.stats_ops:
                self.previous_ops = self.ops
                self.save_stats()
                print "[mcsoda %s] save_stats : %s" %(self.why, latency_cmd)

    def save_stats(self, cur_time=0):
        for key in self.cur:
            if key.startswith('latency-'):
                histo = self.cur.get(key, None)
                if histo:
                    self.sc.latency_stats(key, histo, cur_time)
                    if key.endswith('-recent'):
                        self.cur[key] = {}
        self.sc.sample(self.cur)

    def cmd_append(self, cmd, key_num, key_str, data, expiration, grp):
        self.cmds += 1
        if cmd[0] == 'g' or cmd[0] == 'q':
            hdr, vbucketId = self.header(CMD_GET, key_str, '', opaque=self.cmds)
            m = self.inflight_append_buffer(grp, vbucketId, CMD_GET, self.cmds)
            m.append(hdr)
            m.append(key_str)
            return 1, 0, 0, 0, 0
        elif cmd[0] == 'd':
            hdr, vbucketId = self.header(CMD_DELETE, key_str, '', opaque=self.cmds)
            m = self.inflight_append_buffer(grp, vbucketId, CMD_DELETE, self.cmds)
            m.append(hdr)
            m.append(key_str)
            return 0, 0, 1, 0, 0

        rv = (0, 1, 0, 0, 0)
        curr_cmd = CMD_SET
        curr_extra = struct.pack(SET_PKT_FMT, 0, expiration)

        if cmd[0] == 'a':
            rv = (0, 0, 0, 1, 0)
            curr_cmd, have_extra = self.arpa[self.cur.get('cur-sets', 0) % len(self.arpa)]
            if not have_extra:
                curr_extra = ''

        hdr, vbucketId = self.header(curr_cmd, key_str, data,
                                     extra=curr_extra, opaque=key_num)
        m = self.inflight_append_buffer(grp, vbucketId, curr_cmd, self.cmds)
        m.append(hdr)
        if curr_extra:
            m.append(curr_extra)
        m.append(key_str)
        m.append(data)
        return rv

    def num_ops(self, cur):
        return self.ops

    def recvMsg(self):
        sock = self.conn.s
        buf = self.buf
        pkt, buf = self.readbytes(sock, MIN_RECV_PACKET, buf)
        magic, cmd, keylen, extralen, dtype, errcode, datalen, opaque, cas = \
            struct.unpack(RES_PKT_FMT, pkt)
        if magic != RES_MAGIC_BYTE:
            raise Exception("Unexpected recvMsg magic: " + str(magic))
        val, buf = self.readbytes(sock, datalen, buf)
        self.buf = buf
        if not self.obs_key_cas and cmd == CMD_SET:
            self.obs_key_cas[opaque] = cas  # opaque is the key_num
        if not self.woq_key_cas and cmd == CMD_SET:
            self.woq_key_cas[opaque] = cas
        return cmd, keylen, extralen, errcode, datalen, opaque, val, buf


class StoreMembaseBinary(StoreMemcachedBinary):

    def connect_host_port(self, host, port, user, pswd, bucket="default"):
        """
        Connect to the server host using REST API.
        Username and password should be rest_username and rest_password, \
        generally they are different from ssh identities.
        """
        from membase.api.rest_client import RestConnection
        from memcached.helper.data_helper import VBucketAwareMemcached

        info = { "ip": host, "port": port,
                 'username': user or self.cfg.get("rest_username", "Administrator"),
                 'password': pswd or self.cfg.get("rest_password", "password") }

        self.rest = RestConnection(info)
        self.awareness = VBucketAwareMemcached(self.rest, bucket, info)
        self.backoff = 0
        self.xfer_sent = 0
        self.xfer_recv = 0

    def flush_level(self):
        f = StoreMemcachedBinary.flush_level(self)
        return f * len(self.awareness.memcacheds)

    def inflight_start(self):
        return { 's_bufs': {}, # Key is server str, value is [] of buffer.
                 's_cmds': {}  # Key is server str, value is int (number of cmds).
               }

    def inflight_complete(self, inflight_grp):
        rv = [] # Array of tuples (server, buffer).
        s_bufs = inflight_grp['s_bufs']
        for server in s_bufs.keys():
            buffers = s_bufs[server]
            rv.append((server, ''.join(buffers)))
        return rv

    def inflight_send(self, inflight_msg):
        """
        If timeout value is 0,
        blocks until everything been sent out \
        or the connection breaks.
        """
        timeout_sec = self.cfg.get("socket-timeout", 0)

        sent_total = 0
        for server, buf in inflight_msg:

            length = len(buf)
            if length == 0:
                continue

            sent_tuple = 0   # byte sent out per tuple in inflight_msg
            while sent_tuple < length:
                try:
                    conn = self.awareness.memcacheds[server]
                    if timeout_sec > 0:
                        conn.s.settimeout(timeout_sec)
                    sent = conn.s.send(buf)
                    if sent == 0:
                        log.error("[mcsoda] StoreMembaseBinary.send-zero / skt.send returned 0.")
                        self.cur["cur-StoreMembaseBinary.send-zero"] = self.cur.get("cur-StoreMembaseBinary.send-zero", 0) + 1
                        break
                    sent_tuple += sent
                except socket.timeout:
                    log.error("[mcsoda] EXCEPTION: StoreMembaseBinary.send-socket.timeout / inflight_send timed out")
                    self.cur["cur-ex-StoreMembaseBinary.send-socket.timeout"] = self.cur.get("cur-ex-StoreMembaseBinary.send-socket.timeout", 0) + 1
                    break
                except Exception as e:
                    log.error("[mcsoda] EXCEPTION: StoreMembaseBinary.send / inflight_send: " + str(e))
                    self.cur["cur-ex-StoreMembaseBinary.send"] = self.cur.get("cur-ex-StoreMembaseBinary.send", 0) + 1
                    break

            sent_total += sent_tuple
        return sent_total

    def inflight_recv(self, inflight, inflight_grp, expectBuffer=None):
        received = 0
        s_cmds = inflight_grp['s_cmds']
        reset_my_awareness = False
        backoff = False

        for server in s_cmds.keys():
            try:
                conn = self.awareness.memcacheds[server]
                try:
                    recvBuf = conn.recvBuf
                except:
                    recvBuf = ''
                if expectBuffer == False and recvBuf != '':
                    raise Exception("Was expecting empty buffer, but have (" + \
                                    str(len(recvBuf)) + "): " + recvBuf)
                cmds = s_cmds[server]
                for i in range(cmds):
                    try:
                        rcmd, keylen, extralen, errcode, datalen, ropaque, val, recvBuf = \
                            self.recvMsgSockBuf(conn.s, recvBuf)
                        received += datalen + MIN_RECV_PACKET
                        if errcode == ERR_NOT_MY_VBUCKET:
                            reset_my_awareness = True
                        elif errcode == ERR_ENOMEM or \
                             errcode == ERR_EBUSY or \
                             errcode == ERR_ETMPFAIL:
                            backoff = True
                            log.error("[mcsoda] inflight recv errorcode = %s" %errcode)
                    except Exception as e:
                        log.error("[mcsoda] EXCEPTION: StoreMembaseBinary.recvMsgSockBuf / inflight_recv inner: " + str(e))
                        self.cur["cur-ex-StoreMembaseBinary.recvMsgSockBuf"] = self.cur.get("cur-ex-StoreMembaseBinary.recvMsgSockBuf", 0) + 1
                        reset_my_awareness = True
                        backoff = True
                conn.recvBuf = recvBuf
            except Exception as e:
                log.error("[mcsoda] EXCEPTION: StoreMembaseBinary.inflight_recv / outer: " + str(e))
                self.cur["cur-ex-StoreMembaseBinary.inflight_recv"] = self.cur.get("cur-ex-StoreMembaseBinary.inflight_recv", 0) + 1
                reset_my_awareness = True
                backoff = True

        if backoff:
            self.backoff = max(self.backoff, 0.1) * \
                          self.cfg.get('backoff-factor', 2.0)
            if self.backoff > 0:
                self.cur['cur-backoffs'] = self.cur.get('cur-backoffs', 0) + 1
                log.info("[mcsoda] inflight recv backoff = %s" %self.backoff)
                time.sleep(self.backoff)
        else:
            self.backoff = 0

        if reset_my_awareness:
            try:
                self.awareness.reset()
            except Exception as e:
                log.error("[mcsoda] EXCEPTION: StoreMembaseBinary.awareness.reset: " + str(e))
                self.cur["cur-ex-StoreMembaseBinary.awareness.reset"] = self.cur.get("cur-ex-StoreMembaseBinary.awareness.reset", 0) + 1
                print "EXCEPTION: self.awareness.reset()"
                pass

        return received

    def recvMsgSockBuf(self, sock, buf):
        pkt, buf = self.readbytes(sock, MIN_RECV_PACKET, buf)
        magic, cmd, keylen, extralen, dtype, errcode, datalen, opaque, cas = \
            struct.unpack(RES_PKT_FMT, pkt)
        if magic != RES_MAGIC_BYTE:
            raise Exception("Unexpected recvMsg magic: " + str(magic))
        if not self.obs_key_cas and cmd == CMD_SET:
            self.obs_key_cas[opaque] = cas  # opaque is the key_num
        if not self.woq_key_cas and cmd == CMD_SET:
            self.woq_key_cas[opaque] = cas  # opaque is the key_num
        val, buf = self.readbytes(sock, datalen, buf)
        return cmd, keylen, extralen, errcode, datalen, opaque, val, buf

    def inflight_append_buffer(self, grp, vbucketId, opcode, opaque):
        s_bufs = grp['s_bufs']
        s_cmds = grp['s_cmds']
        s = self.awareness.vBucketMap[vbucketId]
        m = s_bufs.get(s, None)
        if m is None:
            m = []
            s_bufs[s] = m
            s_cmds[s] = 0
        s_cmds[s] += 1
        return m


class StoreMemcachedAscii(Store):

    def connect(self, target, user, pswd, cfg, cur, bucket="default"):
        self.cfg = cfg
        self.cur = cur
        self.target = target
        self.host_port = (target + ":11211").split(':')[0:2]
        self.host_port[1] = int(self.host_port[1])
        self.skt = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.skt.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.skt.connect(tuple(self.host_port))
        self.queue = []
        self.ops = 0
        self.previous_ops = 0
        self.buf = ''
        self.arpa = [ 'add', 'replace', 'append', 'prepend' ]
        self.xfer_sent = 0
        self.xfer_recv = 0

    def command(self, c):
        self.queue.append(c)
        if len(self.queue) > (self.cur.get('batch') or \
                              self.cfg.get('batch', 100)):
            self.flush()
            return True
        return False

    def command_send(self, cmd, key_num, key_str, data, expiration):
        if cmd[0] == 'g' or cmd[0] == 'q':
            return cmd + ' ' + key_str + '\r\n'
        if cmd[0] == 'd':
            return 'delete ' + key_str + '\r\n'

        c = 'set'
        if cmd[0] == 'a':
           c = self.arpa[self.cur.get('cur-sets', 0) % len(self.arpa)]
        return "%s %s 0 %s %s\r\n%s\r\n" % (c, key_str, expiration,
                                            len(data), data)

    def command_recv(self, cmd, key_num, key_str, data, expiration):
        buf = self.buf
        if cmd[0] == 'g' or cmd[0] == 'q':
            # GET...
            line, buf = self.readline(self.skt, buf)
            while line and line != 'END':
                # line == "VALUE k flags len"
                rvalue, rkey, rflags, rlen = line.split()
                data, buf = self.readbytes(self.skt, int(rlen) + 2, buf)
                line, buf = self.readline(self.skt, buf)
        elif cmd[0] == 'd':
            # DELETE...
            line, buf = self.readline(self.skt, buf) # line == "DELETED"
        else:
            # SET...
            line, buf = self.readline(self.skt, buf) # line == "STORED"
        self.buf = buf

    def flush(self):
        m = []
        for c in self.queue:
            cmd, key_num, key_str, data, expiration = c
            m.append(self.command_send(cmd, key_num, key_str, data, expiration))

        self.skt.send(''.join(m))

        for c in self.queue:
            cmd, key_num, key_str, data, expiration = c
            self.command_recv(cmd, key_num, key_str, data, expiration)

        self.ops += len(self.queue)
        self.queue = []

    def num_ops(self, cur):
        return self.ops

    def readline(self, skt, buf):
        while True:
            index = buf.find('\r\n')
            if index >= 0:
                break
            data = skt.recv(4096)
            if not data:
                return '', ''
            buf += data
        return buf[:index], buf[index+2:]

# --------------------------------------------------------

# A key is a 16 char hex string.
def key_to_name(key_num, key_str):
    return "%s %s" % (key_str[-16:-12], key_str[-4:-1])

def key_to_email(key_num, key_str):
    return "%s@%s.com" % (key_str[-16:-12], key_str[-13:-11])

def key_to_city(key_num, key_str):
    return key_str[-12:-9]

def key_to_country(key_num, key_str):
    return key_str[-9:-7]

def key_to_realm(key_num, key_str):
    return key_str[-7:-5]

def key_to_coins(key_num, key_str):
    sub_key = key_str[-16:]
    return max(0.0, int(sub_key[0:4], 16) / 100.0)

def key_to_category(key_num, key_str):
    return int(key_str[-12], 16) % 3

def key_to_achievements(key_num, key_str):
    next = 300
    achievements = []
    sub_key = key_str[-16:]
    for i in range(len(sub_key)):
        next = (next + int(sub_key[i], 16) * i) % 500
        if next < 256:
            achievements.append(next)
    return achievements

doc_cache = {}

def gen_doc_string(key_num, key_str, min_value_size, suffix, json,
                   cache=None, key_name="key", suffix_ex="", whitespace=True):
    global doc_cache

    c = "{"
    if not json:
        c = "*"

    d = None
    if cache:
        d = doc_cache.get(key_num, None)

    if d is None:
        d = """"%s":"%s",
 "key_num":%s,
 "name":"%s",
 "email":"%s",
 "city":"%s",
 "country":"%s",
 "realm":"%s",
 "coins":%s,
 "category":%s,
 "achievements":%s,""" % (key_name, key_str,
                          key_num,
                          key_to_name(key_num, key_str),
                          key_to_email(key_num, key_str),
                          key_to_city(key_num, key_str),
                          key_to_country(key_num, key_str),
                          key_to_realm(key_num, key_str),
                          key_to_coins(key_num, key_str),
                          key_to_category(key_num, key_str),
                          key_to_achievements(key_num, key_str))
        if not whitespace:
            d = d.replace("\n ", "")
        if cache:
            doc_cache[key_num] = d

    return "%s%s%s%s" % (c, d, suffix_ex, suffix)

# --------------------------------------------------------

PROTOCOL_STORE = { 'memcached-ascii': StoreMemcachedAscii,
                   'memcached-binary': StoreMemcachedBinary,
                   'membase-binary': StoreMembaseBinary,
                   'none-binary': Store,
                   'none': Store }

def run(cfg, cur, protocol, host_port, user, pswd,
        stats_collector = None, stores = None, ctl = None, heartbeat = 0, why = "", bucket = "default"):
    if type(cfg['min-value-size']) == type(""):
        cfg['min-value-size'] = string.split(cfg['min-value-size'], ",")
    if type(cfg['min-value-size']) != type([]):
        cfg['min-value-size'] = [ cfg['min-value-size'] ]

    cfg['body'] = {}
    cfg['suffix'] = {}

    for i in range(len(cfg['min-value-size'])):
        mvs = int(cfg['min-value-size'][i])
        cfg['min-value-size'][i] = mvs
        cfg['body'][mvs] = 'x'
        while len(cfg['body'][mvs]) < mvs:
            cfg['body'][mvs] = cfg['body'][mvs] + \
                               md5(str(len(cfg['body'][mvs]))).hexdigest()
        cfg['suffix'][mvs] = "\"body\":\"" + cfg['body'][mvs] + "\"}"

    ctl = ctl or { 'run_ok': True }

    threads = []

    for i in range(cfg.get('threads', 1)):
        store = None
        if stores and i < len(stores):
            store = stores[i]

        if store is None:
            store = PROTOCOL_STORE[protocol]()

        log.info("store: %s - %s" % (i, store.__class__))

        store.connect(host_port, user, pswd, cfg, cur, bucket=bucket)
        store.stats_collector(stats_collector)

        threads.append(threading.Thread(target=run_worker,
                                       args=(ctl, cfg, cur, store,
                                             "thread-" + str(i) + ": ")))

    store.show_some_keys()

    if cfg.get("doc-cache", 0) > 0 and cfg.get("doc-gen", 0) > 0:
        min_value_size = cfg['min-value-size'][0]
        json = cfg.get('json', 1) > 0
        cache = cfg.get('doc-cache', 0)
        log.info("doc-gen...")
        gen_start = time.time()
        for key_num in range(cfg.get("max-items", 0)):
            key_str = prepare_key(key_num, cfg.get('prefix', ''))
            store.gen_doc(key_num, key_str, min_value_size, json, cache)
        gen_end = time.time()
        log.info("doc-gen...done (elapsed: %s, docs/sec: %s)" % \
                 (gen_end - gen_start,
                  float(key_num) / (gen_end - gen_start)))

    def stop_after(secs):
        time.sleep(secs)
        ctl['run_ok'] = False

    if cfg.get('time', 0) > 0:
        t = threading.Thread(target=stop_after, args=(cfg.get('time', 0),))
        t.daemon = True
        t.start()

    t_start = time.time()

    try:
        if len(threads) <= 1:
            run_worker(ctl, cfg, cur, store, "", heartbeat, why)
        else:
            for thread in threads:
                thread.daemon = True
                thread.start()

            while len(threads) > 0:
                threads[0].join(1)
                threads = [t for t in threads if t.isAlive()]
    except KeyboardInterrupt:
        ctl['run_ok'] = False

    t_end = time.time()

    # Final stats
    log.info("")
    log.info(dict_to_s(cur))
    total_time = float(t_end - t_start)
    if cur.get('cur-queries', 0):
        total_cmds = cur.get('cur-queries', 0)
    else:
        total_cmds = cur.get('cur-gets', 0) + cur.get('cur-sets', 0)
    log.info("    ops/sec: %s" %(total_cmds / total_time))

    threads = [t for t in threads if t.isAlive()]
    heartbeat = 0
    while len(threads) > 0:
        threads[0].join(1)
        heartbeat = heartbeat + 1
        if heartbeat >= 60:
            heartbeat = 0
            log.info("    mcsoda is running with %s threads" % len(threads))
        threads = [t for t in threads if t.isAlive()]

    ctl['run_ok'] = False
    if ctl.get('shutdown_event') is not None:
        ctl['shutdown_event'].set()
    log.info("[mcsoda: %s] stopped running." %why)
    return cur, t_start, t_end

# --------------------------------------------------------

def main(argv, cfg_defaults=None, cur_defaults=None, protocol=None, stores=None,
         extra_examples=None):
    cfg_defaults = cfg_defaults or {
        "prefix":             ("",    "Prefix for every item key."),
        "max-ops":            (0,     "Max # of ops before exiting. 0 means keep going."),
        "max-items":          (-1,    "Max # of items; default 100000."),
        "max-creates":        (-1,    "Max # of creates; defaults to max-items."),
        "min-value-size":     ("10",  "Min value size (bytes) for SET's; comma-separated."),
        "ratio-sets":         (0.1,   "Fraction of requests that should be SET's."),
        "ratio-creates":      (0.1,   "Fraction of SET's that should create new items."),
        "ratio-misses":       (0.05,  "Fraction of GET's that should miss."),
        "ratio-hot":          (0.2,   "Fraction of items to have as a hot item subset."),
        "ratio-hot-sets":     (0.95,  "Fraction of SET's that hit the hot item subset."),
        "ratio-hot-gets":     (0.95,  "Fraction of GET's that hit the hot item subset."),
        "ratio-deletes":      (0.0,   "Fraction of SET updates that shold be DELETE's."),
        "ratio-arpas":        (0.0,   "Fraction of SET non-DELETE'S to be 'a-r-p-a' cmds."),
        "ratio-expirations":  (0.0,   "Fraction of SET's that use the provided expiration."),
        "ratio-queries":      (0.0,   "Fraction of GET hits that should be queries."),
        "expiration":         (0,     "Expiration time parameter for SET's"),
        "exit-after-creates": (0,     "Exit after max-creates is reached."),
        "threads":            (1,     "Number of client worker threads to use."),
        "batch":              (100,   "Batch/pipeline up this # of commands per server."),
        "json":               (1,     "Use JSON documents. 0 to generate binary documents."),
        "time":               (0,     "Stop after this many seconds if > 0."),
        "max-ops-per-sec":    (0,     "When >0, max ops/second target performance."),
        "report":             (40000, "Emit performance output after this many requests."),
        "histo-precision":    (1,     "Precision of histogram bins."),
        "vbuckets":           (0,     "When >0, vbucket hash in memcached-binary protocol."),
        "doc-cache":          (1,     "When 1, cache docs; faster, but uses O(N) memory."),
        "doc-gen":            (1,     "When 1 and doc-cache, pre-generate docs at start."),
        "backoff-factor":     (2.0,   "Exponential backoff factor on ETMPFAIL errors."),
        "hot-shift":          (0,     "# of keys/sec that hot item subset should shift."),
        "random":             (0,     "When 1, use random keys for gets and updates."),
        "queries":            ("",    "Query templates; semicolon-separated."),
        "socket-timeout":     (0,     "Used for socket.settimeout(), in seconds.")}

    cur_defaults = cur_defaults or {
        "cur-items":    (0, "Number of items known to already exist."),
        "cur-sets":     (0, "Number of sets already done."),
        "cur-creates":  (0, "Number of sets that were creates."),
        "cur-gets":     (0, "Number of gets already done."),
        "cur-deletes":  (0, "Number of deletes already done."),
        "cur-arpas":    (0, "# of add/replace/prepend/append's (a-r-p-a) cmds."),
        "cur-queries":  (0, "Number of gets that were view/index queries."),
        "cur-base":     (0, "Base of numeric key range. 0 by default.")}

    if len(argv) < 2 or "-h" in argv or "--help" in argv:
        print("usage: %s [memcached[-binary|-ascii]://][user[:pswd]@]host[:port] [key=val]*\n" %
              (argv[0]))
        print("  default protocol = memcached-binary://")
        print("  default port     = 11211\n")
        examples = ["examples:",
                    "  %s membase://127.0.0.1:8091 max-items=1000000 json=1",
                    "  %s memcached://127.0.0.1:11210 vbuckets=1024",
                    "  %s memcached://127.0.0.1:11211",
                    "  %s memcached-ascii://127.0.0.1:11211",
                    "  %s memcached-binary://127.0.0.1:11211",
                    "  %s 127.0.0.1:11211",
                    "  %s 127.0.0.1",
                    "  %s my-test-bucket@127.0.0.1",
                    "  %s my-test-bucket:MyPassword@127.0.0.1",
                    "  %s none://"]
        if extra_examples:
            examples = examples + extra_examples
        for s in examples:
            if s.find("%s") > 0:
                print(s % (argv[0]))
            else:
                print(s)
        print("")
        print("optional key=val's and their defaults:")
        for d in [cfg_defaults, cur_defaults]:
            for k in sorted(d.iterkeys()):
                print("  %s = %s %s" %
                      (string.ljust(k, 18), string.ljust(str(d[k][0]), 5), d[k][1]))
        print("")
        print("  TIP: min-value-size can be comma-separated values: min-value-size=10,256,1024")
        print("")
        sys.exit(-1)

    cfg = {}
    cur = {}
    err = {}

    for (o, d) in [(cfg, cfg_defaults), (cur, cur_defaults)]: # Parse key=val pairs.
        for (dk, dv) in d.iteritems():
            o[dk] = dv[0]
        for kv in argv[2:]:
            s = (kv + '=').split('=')[0:-1]
            k = s[0]
            v = '='.join(s[1:])
            if k and v and k in o:
                if type(o[k]) != type(""):
                    try:
                        v = ({ 'y':'1', 'n':'0' }).get(v, v)
                        for parse in [float, int]:
                            if str(parse(v)) == v:
                                v = parse(v)
                    except:
                        err[kv] = err.get(kv, 0) + 1
                o[k] = v
            else:
                err[kv] = err.get(kv, 0) + 1

    for kv in err:
        if err[kv] > 1:
            log.error("problem parsing key=val option: " + kv)
    for kv in err:
        if err[kv] > 1:
            sys.exit(-1)

    if cfg.get('max-items', 0) < 0 and cfg.get('max-creates', 0) < 0:
        cfg['max-items'] = 100000
    if cfg.get('max-items', 0) < 0:
        cfg['max-items'] = cfg.get('max-creates', 0)
    if cfg.get('max-creates', 0) < 0:
        cfg['max-creates'] = cfg.get('max-items', 0)

    for o in [cfg, cur]:
        for k in sorted(o.iterkeys()):
            log.info("    %s = %s" % (string.ljust(k, 20), o[k]))

    protocol = protocol or '-'.join(((["memcached"] + \
                                    argv[1].split("://"))[-2] + "-binary").split('-')[0:2])
    host_port = ('@' + argv[1].split("://")[-1]).split('@')[-1] + ":11211"
    user, pswd = (('@' + argv[1].split("://")[-1]).split('@')[-2] + ":").split(':')[0:2]

    cfg["timing-suffixes"] = [""]

    run(cfg, cur, protocol, host_port, user, pswd, stores=stores)

if __name__ == "__main__":
  main(sys.argv)




from random import shuffle
import time
import logger
from membase.api.exception import StatsUnavailableException, \
    ServerAlreadyJoinedException, RebalanceFailedException
from membase.api.rest_client import RestConnection, RestHelper
from membase.helper.bucket_helper import BucketOperationHelper
from memcached.helper.data_helper import MemcachedClientHelper, VBucketAwareMemcached
from mc_bin_client import MemcachedClient, MemcachedError

log = logger.Logger.get_logger()


class RebalanceHelper():
    @staticmethod
    #bucket is a json object that contains name,port,password
    def wait_for_mc_stats_all_nodes(master, bucket, stat_key, stat_value, timeout_in_seconds=120, verbose=True):
        log.info("waiting for bucket {0} stat : {1} to match {2} on {3}".format(bucket, stat_key, \
                                                                                stat_value, master.ip))
        time_to_timeout = 0
        previous_stat_value = -1
        curr_stat_value = -1
        verified = False
        all_stats = {}
        while not verified:
            rest = RestConnection(master)
            nodes = rest.node_statuses()
            for node in nodes:
                _server = {"ip": node.ip, "port": node.port, "username": master.rest_username,
                           "password": master.rest_password}
                #failed over node is part of node_statuses but since its failed over memcached connections
                #to this node will fail
                node_self = RestConnection(_server).get_nodes_self()
                if node_self.clusterMembership == 'active':
                    mc = MemcachedClientHelper.direct_client(_server, bucket)
                    n_stats = mc.stats("")
                    mc.close()
                    all_stats[node.id] = n_stats
            actual_stat_value = -1
            for k in all_stats:
                if all_stats[k] and stat_key in all_stats[k]:
                    if actual_stat_value == -1:
                        print all_stats[k][stat_key]
                        actual_stat_value = int(all_stats[k][stat_key])
                    else:
                        actual_stat_value += int(all_stats[k][stat_key])
            if actual_stat_value == stat_value:
                log.info("{0} : {1}".format(stat_key, actual_stat_value))
                verified = True
                break
            else:
                if verbose:
                    log.info("{0} : {1}".format(stat_key, actual_stat_value))
                curr_stat_value = actual_stat_value

                # values are changing so clear any timeout
                if curr_stat_value != previous_stat_value:
                    time_to_timeout = 0
                else:
                    if time_to_timeout == 0:
                        time_to_timeout = time.time() + timeout_in_seconds
                    if time_to_timeout < time.time():
                        log.info("no change in {0} stat after {1} seconds (value = {2})".format(stat_key, timeout_in_seconds, curr_stat_value))
                        break

                previous_stat_value = curr_stat_value

                if not verbose:
                    time.sleep(0.1)
                else:
                    time.sleep(2)
        return verified

    @staticmethod
    #bucket is a json object that contains name,port,password
    def wait_for_stats(master, bucket, stat_key, stat_value, timeout_in_seconds=120, verbose=True):
        log.info("waiting for bucket {0} stat : {1} to match {2} on {3}".format(bucket, stat_key, \
                                                                                stat_value, master.ip))
        time_to_timeout = 0
        previous_stat_value = -1
        curr_stat_value = -1
        verified = False
        while not verified:
            rest = RestConnection(master)
            try:
                stats = rest.get_bucket_stats(bucket)
                if stats and stat_key in stats and stats[stat_key] == stat_value:
                    log.info("{0} : {1}".format(stat_key, stats[stat_key]))
                    verified = True
                    break
                else:
                    if stats and stat_key in stats:
                        if verbose:
                            log.info("{0} : {1}".format(stat_key, stats[stat_key]))
                        curr_stat_value = stats[stat_key]

                    # values are changing so clear any timeout
                    if curr_stat_value != previous_stat_value:
                        time_to_timeout = 0
                    else:
                        if time_to_timeout == 0:
                            time_to_timeout = time.time() + timeout_in_seconds
                        if time_to_timeout < time.time():
                            log.info("no change in {0} stat after {1} seconds (value = {2})".format(stat_key, timeout_in_seconds, curr_stat_value))
                            break

                    previous_stat_value = curr_stat_value

                    if not verbose:
                        time.sleep(0.1)
                    else:
                        time.sleep(2)
            except:
                log.info("unable to collect stats from server {0}".format(master))
                verified = True #TODO: throw ex and assume caller catches
                break

        return verified

    @staticmethod
    def wait_for_stats_no_timeout(master, bucket, stat_key, stat_value, timeout_in_seconds=-1, verbose=True):
        log.info("waiting for bucket {0} stat : {1} to match {2} on {3}".format(bucket, stat_key, \
                                                                                stat_value, master.ip))
        rest = RestConnection(master)
        stats = rest.get_bucket_stats(bucket)

        while stats.get(stat_key, -1) != stat_value:
            stats = rest.get_bucket_stats(bucket)
            if verbose:
                log.info("{0} : {1}".format(stat_key, stats.get(stat_key, -1)))
            time.sleep(5)
        return True

    @staticmethod
    #bucket is a json object that contains name,port,password
    def wait_for_mc_stats(master, bucket, stat_key, stat_value, timeout_in_seconds=120, verbose=True):
        log.info("waiting for bucket {0} stat : {1} to match {2} on {3}".format(bucket, stat_key, \
                                                                                stat_value, master.ip))
        start = time.time()
        verified = False
        while (time.time() - start) <= timeout_in_seconds:
            c = MemcachedClient(master.ip, 11210)
            stats = c.stats()
            c.close()
            if stats and stat_key in stats and str(stats[stat_key]) == str(stat_value):
                log.info("{0} : {1}".format(stat_key, stats[stat_key]))
                verified = True
                break
            else:
                if stats and stat_key in stats:
                    if verbose:
                        log.info("{0} : {1}".format(stat_key, stats[stat_key]))
                if not verbose:
                    time.sleep(0.1)
                else:
                    time.sleep(2)
        return verified

    @staticmethod
    def wait_for_mc_stats_no_timeout(master, bucket, stat_key, stat_value, timeout_in_seconds=-1, verbose=True):
        log.info("waiting for bucket {0} stat : {1} to match {2} on {3}".format(bucket, stat_key, \
                                                                                stat_value, master.ip))
        # keep retrying until reaches the server
        stats = {}
        while not stats:
            try:
                c = MemcachedClient(master.ip, 11210)
                c.sasl_auth_plain(bucket, '')
                stats = c.stats()
            except Exception as e:
                log.info("Exception: {0}, retry in 2 seconds ...".format(str(e)))
                stats = {}
                time.sleep(2)
            finally:
                c.close()

        while str(stats[stat_key]) != str(stat_value):
            c = MemcachedClient(master.ip, 11210)
            c.sasl_auth_plain(bucket, '')
            stats = c.stats()
            c.close()
            if verbose:
                log.info("{0} : {1}".format(stat_key, stats[stat_key]))
            time.sleep(5)
        return True

    @staticmethod
    #bucket is a json object that contains name,port,password
    def wait_for_stats_int_value(master, bucket, stat_key, stat_value, option="==",timeout_in_seconds=120, verbose=True):
        log.info("waiting for bucket {0} stat : {1} to {2} {3} on {4}".format(bucket, stat_key,option, \
                                                                                stat_value, master.ip))
        start = time.time()
        verified = False
        while (time.time() - start) <= timeout_in_seconds:
            rest = RestConnection(master)
            stats = rest.get_bucket_stats(bucket)
            #some stats are in memcached
            if stats and stat_key in stats:
                actual = int(stats[stat_key])
                if option == "==":
                    verified = stat_value == actual
                elif option == ">":
                    verified = stat_value > actual
                elif option == "<":
                    verified = stat_value < actual
                elif option == ">=":
                    verified = stat_value >= actual
                elif option == "<=":
                    verified = stat_value <= actual
                if verified:
                    log.info("verified {0} : {1}".format(stat_key, actual))
                    break
                if verbose:
                    log.info("{0} : {1} isn't {2} {3}".format(stat_key, stat_value, option, actual))
            time.sleep(2)
        return verified

    @staticmethod
    #bucket is a json object that contains name,port,password
    def wait_for_stats_on_all(master, bucket, stat_key, stat_value, timeout_in_seconds=120,
                              fn=None):
        fn = fn or RebalanceHelper.wait_for_stats
        rest = RestConnection(master)
        servers = rest.get_nodes()
        verified = False
        start_time = time.time()
        for server in servers:
            verified = fn(server, bucket, stat_key, stat_value, \
                          timeout_in_seconds=timeout_in_seconds)
            if not verified:
                log.info("bucket {0}: stat_key {1} for server {2} timed out in {3}".format(bucket, stat_key, \
                                                                                           server.ip, time.time()-start_time))
                break

        return verified

    @staticmethod
    def wait_till_total_numbers_match(master,
                                      bucket,
                                      timeout_in_seconds=120):

        log.info('waiting for sum_of_curr_items == total_items....')
        start = time.time()
        verified = False
        while (time.time() - start) <= timeout_in_seconds:
            try:
                if RebalanceHelper.verify_items_count(master, bucket):
                    verified = True
                    break
                else:
                    time.sleep(2)
            except StatsUnavailableException:
                log.info("unable to retrieve stats for any node. returning true")
                break
        rest = RestConnection(master)
        RebalanceHelper.print_taps_from_all_nodes(rest, bucket)
        return verified

    @staticmethod
    def wait_for_persistence(master, bucket, timeout=120):
        RebalanceHelper.wait_for_mc_stats_all_nodes(
            master, bucket, "ep_queue_size", 0,
            timeout_in_seconds=timeout)
        RebalanceHelper.wait_for_mc_stats_all_nodes(
            master, bucket, "ep_flusher_todo", 0,
            timeout_in_seconds=timeout)
        RebalanceHelper.wait_for_mc_stats_all_nodes(
            master, bucket, "ep_uncommitted_items", 0,
            timeout_in_seconds=timeout)

    @staticmethod
    #TODO: add password and port
    def print_taps_from_all_nodes(rest, bucket='default'):
        #get the port number from rest ?

        log = logger.Logger.get_logger()
        nodes_for_stats = rest.get_nodes()
        for node_for_stat in nodes_for_stats:
            try:
                client = MemcachedClientHelper.direct_client(node_for_stat, bucket)
                log.info("getting tap stats.. for {0}".format(node_for_stat.ip))
                tap_stats = client.stats('tap')
                if tap_stats:
                    RebalanceHelper.log_interesting_taps(node_for_stat, tap_stats, log)
                client.close()
            except Exception as ex:
                log.error("error {0} while getting stats...".format(ex))

    @staticmethod
    def log_interesting_taps(node, tap_stats, logger):
        interesting_stats = ['ack_log_size', 'ack_seqno', 'ack_window_full', 'has_item', 'has_queued_item',
                             'idle', 'paused', 'pending_backfill', 'pending_disk_backfill', 'recv_ack_seqno',
                             'ep_num_new_']
        for name in tap_stats:
            for interesting_stat in interesting_stats:
                if name.find(interesting_stat) != -1:
                    logger.info("TAP {0} :{1}   {2}".format(node.id, name, tap_stats[name]))
                    break

    @staticmethod
    def verify_items_count(master, bucket, num_attempt=3, timeout=2):
        #get the #of buckets from rest
        rest = RestConnection(master)
        bucket_info = rest.get_bucket(bucket, num_attempt, timeout)
        replica_factor = bucket_info.numReplicas
        vbucket_active_sum = 0
        vbucket_replica_sum = 0
        vbucket_pending_sum = 0
        all_server_stats = []
        stats_received = True
        nodes = rest.get_nodes()
        for server in nodes:
            #get the stats
            server_stats = rest.get_bucket_stats_for_node(bucket, server)
            if not server_stats:
                log.info("unable to get stats from {0}:{1}".format(server.ip, server.port))
                stats_received = False
            all_server_stats.append((server, server_stats))
        if not stats_received:
            raise StatsUnavailableException()
        sum = 0
        for server, single_stats in all_server_stats:
            if not single_stats or "curr_items" not in single_stats:
                continue
            sum += single_stats["curr_items"]
            log.info("curr_items from {0}:{1} : {2}".format(server.ip, server.port, \
                single_stats["curr_items"]))
            if 'vb_pending_num' in single_stats:
                vbucket_pending_sum += single_stats['vb_pending_num']
                log.info(
                    "vb_pending_num from {0}:{1} : {2}".format(server.ip, server.port, \
                        single_stats["vb_pending_num"]))
            if 'vb_active_num' in single_stats:
                vbucket_active_sum += single_stats['vb_active_num']
                log.info(
                    "vb_active_num from {0}:{1} : {2}".format(server.ip, server.port, \
                        single_stats["vb_active_num"]))
            if 'vb_replica_num' in single_stats:
                vbucket_replica_sum += single_stats['vb_replica_num']
                log.info(
                    "vb_replica_num from {0}:{1} : {2}".format(server.ip, server.port, \
                        single_stats["vb_replica_num"]))

        msg = "summation of vb_active_num : {0} vb_pending_num : {1} vb_replica_num : {2}"
        log.info(msg.format(vbucket_active_sum, vbucket_pending_sum, vbucket_replica_sum))
        msg = 'sum : {0} and sum * (replica_factor + 1) ({1}) : {2}'
        log.info(msg.format(sum, replica_factor + 1, (sum * (replica_factor + 1))))
        master_stats = rest.get_bucket_stats(bucket)
        log.info('master_stats : {0}'.format(master_stats["curr_items_tot"]))
        delta = sum * (replica_factor + 1) - master_stats["curr_items_tot"]
        delta = abs(delta)

        if delta > 0:
            if sum==0:
                missing_percentage = 0
            else:
                missing_percentage = delta * 1.0 / (sum * (replica_factor + 1))
        else:
            missing_percentage = 1
        log.info("delta : {0} missing_percentage : {1} replica_factor : {2}".format(delta, \
            missing_percentage, replica_factor))
        # If no items missing then, return True
        if not delta:
            return True
        return False

    @staticmethod
    def verify_maps(vbucket_map_before, vbucket_map_after):
        #for each bucket check the replicas
        for i in range(0, len(vbucket_map_before)):
            if not vbucket_map_before[i].master == vbucket_map_after[i].master:
                log.error(
                    'vbucket[{0}].master mismatch {1} vs {2}'.format(i, vbucket_map_before[i].master,
                                                                     vbucket_map_after[i].master))
                return False
            for j in range(0, len(vbucket_map_before[i].replica)):
                if not (vbucket_map_before[i].replica[j]) == (vbucket_map_after[i].replica[j]):
                    log.error('vbucket[{0}].replica[{1} mismatch {2} vs {3}'.format(i, j,
                                                                                    vbucket_map_before[i].replica[j],
                                                                                    vbucket_map_after[i].replica[j]))
                    return False
        return True

    #read the current nodes
    # if the node_ip already added then just
    #silently return
    #if its not added then let try to add this and then rebalance
    #we should alo try to get the bucket information from
    #rest api instead of passing it to the fucntions

    @staticmethod
    def rebalance_in(servers, how_many, do_shuffle=True, monitor=True, do_check=True):
        servers_rebalanced = []
        log = logger.Logger.get_logger()
        rest = RestConnection(servers[0])
        nodes = rest.node_statuses()
        #are all ips the same
        nodes_on_same_ip = True
        firstIp = nodes[0].ip
        if len(nodes) == 1:
            nodes_on_same_ip  = False
        else:
            for node in nodes:
                if node.ip != firstIp:
                    nodes_on_same_ip = False
                    break
        nodeIps = ["{0}:{1}".format(node.ip,node.port) for node in nodes]
        log.info("current nodes : {0}".format(nodeIps))
        toBeAdded = []
        master = servers[0]
        selection = servers[1:]
        if do_shuffle:
            shuffle(selection)
        for server in selection:
            if nodes_on_same_ip:
                if not "{0}:{1}".format(firstIp,server.port) in nodeIps:
                    toBeAdded.append(server)
                    servers_rebalanced.append(server)
                    log.info("choosing {0}:{1}".format(server.ip, server.port))
            elif not "{0}:{1}".format(server.ip,server.port) in nodeIps:
                toBeAdded.append(server)
                servers_rebalanced.append(server)
                log.info("choosing {0}:{1}".format(server.ip, server.port))
            if len(toBeAdded) == int(how_many):
                break

        if do_check and len(toBeAdded) < how_many:
            raise Exception("unable to find {0} nodes to rebalance_in".format(how_many))

        for server in toBeAdded:
            otpNode = rest.add_node(master.rest_username, master.rest_password,
                                    server.ip, server.port)
        otpNodes = [node.id for node in rest.node_statuses()]
        started = rest.rebalance(otpNodes, [])
        msg = "rebalance operation started ? {0}"
        log.info(msg.format(started))
        if monitor is not True:
            return True, servers_rebalanced
        if started:
            try:
                result = rest.monitorRebalance()
            except RebalanceFailedException as e:
                log.error("rebalance failed: {0}".format(e))
                return False, servers_rebalanced
            msg = "successfully rebalanced in selected nodes from the cluster ? {0}"
            log.info(msg.format(result))
            return result, servers_rebalanced
        return False, servers_rebalanced

    @staticmethod
    def begin_rebalance_in(master, servers, timeout=5):
        log = logger.Logger.get_logger()
        rest = RestConnection(master)
        otpNode = None

        for server in servers:
            if server == master:
                continue
            log.info("adding node {0}:{1} to cluster".format(server.ip, server.port))
            try:
                otpNode = rest.add_node(master.rest_username, master.rest_password, server.ip, server.port)
                msg = "unable to add node {0}:{1} to the cluster"
                assert otpNode, msg.format(server.ip, server.port)
            except ServerAlreadyJoinedException:
                log.info("server {0} already joined".format(server))
        log.info("beginning rebalance in")
        try:
            rest.rebalance(otpNodes=[node.id for node in rest.node_statuses()], ejectedNodes=[])
        except:
            log.error("rebalance failed, trying again after {0} seconds".format(timeout))

    @staticmethod
    def begin_rebalance_out(master, servers, timeout=5):
        log = logger.Logger.get_logger()
        rest = RestConnection(master)

        master_node = rest.get_nodes_self()

        allNodes = []
        ejectedNodes = []
        nodes = rest.node_statuses()
        for server in servers:
            server_node = RestConnection(server).get_nodes_self()
            if server_node == master_node:
                continue
            log.info("removing node {0}:{1} from cluster".format(server_node.ip, server_node.port))
            for node in nodes:
                if "{0}:{1}".format(node.ip, node.port) == "{0}:{1}".format(server_node.ip, server_node.port):
                    ejectedNodes.append(node.id)
        log.info("beginning rebalance out")
        try:
            rest.rebalance(otpNodes=[node.id for node in nodes], ejectedNodes=ejectedNodes)
        except:
            log.error("rebalance failed, trying again after {0} seconds".format(timeout))

    @staticmethod
    def end_rebalance(master):
        log = logger.Logger.get_logger()
        rest = RestConnection(master)
        assert rest.monitorRebalance(), "rebalance operation failed after adding nodes"
        log.info("rebalance finished")

    @staticmethod
    def getOtpNodeIds(master):
        rest = RestConnection(master)
        nodes = rest.node_statuses()
        otpNodeIds = [node.id for node in nodes]
        return otpNodeIds

    @staticmethod
    def verify_vBuckets_info(master, bucket="default"):
        '''
        verify vBuckets' state and items count(for active/replica) in them related to vBucketMap for all nodes in cluster
        '''
        awareness = VBucketAwareMemcached(RestConnection(master), bucket)
        vb_map=awareness.vBucketMap
        vb_mapReplica=awareness.vBucketMapReplica
        replica_num=len(vb_mapReplica[0])

        #get state and count items for all vbuckets for each node
        node_stats=RebalanceHelper.get_vBuckets_info(master)
        state=True
        #iterate throught all vbuckets by their numbers
        for num in vb_map:
            #verify that active vbucket in memcached  is also active in stats("hash)
            if(node_stats[vb_map[num]]["vb_"+ str(num)][0]!="active"):
                log.info("vBucket {0} in {1} node has wrong state {3}".format("vb_"+ str(num), vb_map[num], node_stats[vb_map[num]]["vb_"+ str(num)]));
                state = False
            #number of active items for num vBucket
            vb = node_stats[vb_map[num]]["vb_"+ str(num)][1]
            active_vb=vb_map[num]
            #list of nodes for wich num vBucket is replica
            replica_vbs=vb_mapReplica[key]
            sum_items_replica=0
            #sum of replica items for all nodes for num vBucket
            for i in range(replica_num):
                if(node_stats[vb_mapReplica[num][i]]["vb_"+ str(num)][0]!="replica"):
                    log.info("vBucket {0} in {1} node has wrong state {3}".format("vb_"+ str(num), vb_mapReplica[num], node_stats[vb_mapReplica[num]]["vb_"+ str(num)]));
                    state = False
                sum_items_replica+=int(node_stats[replica_vbs[i]]["vb_"+ str(num)][1])
            #print information about the discrepancy of the number of replica and active items for num vBucket
            if (int(vb)*len(vb_mapReplica[num])!=sum_items_replica):
                log.info("sum of active items doesn't correspond to replica's vBucets in {0} vBucket:".format("vb_"+ str(num)))
                log.info("items in active vBucket {0}:{1}".format(vb_map[num], node_stats[vb_map[num]]["vb_"+ str(num)]))
                for j in range(replica):
                    log.info("items in replica vBucket {0}: {1}".format(vb_mapReplica[num][j], node_stats[vb_mapReplica[num][j]]["vb_"+ str(num)]))
                    log.info(node_stats[vb_mapReplica[num][0]])
                state = False

        if not state:
             log.error("Something is wrong, see log above. See details:")
             log.error("vBucetMap: {0}".format(vb_map))
             log.error("vBucetReplicaMap: {0}".format(vb_mapReplica))
             log.error("node_stats: {0}".format(node_stats))
        return state

    @staticmethod
    def get_vBuckets_info(master):
        """
        return state and count items for all vbuckets for each node
        format: dict: {u'1node_ip1': {'vb_79': ['replica', '0'], 'vb_78': ['active', '0']..}, u'1node_ip1':....}
        """
        rest = RestConnection(master)
        port = rest.get_nodes_self().memcached
        nodes = rest.node_statuses()
        _nodes_stats= {}
        for node in nodes:
            stat={}
            buckets = []
            _server = {"ip": node.ip, "port": node.port, "username": master.rest_username,
                           "password": master.rest_password}
            try:
                buckets = rest.get_buckets()
                mc = MemcachedClient(node.ip, port)
                stat_hash = mc.stats("hash")
            except Exception:
                if not buckets:
                    log.error("There are not any buckets in {0}:{1} node".format(node.ip, node.port))
                else:
                    log.error("Impossible to get vBucket's information for {0}:{1} node".format(node.ip, node.port))
                    _nodes_stats[node.ip+":"+str(node.port)]
                continue
            mc.close()
            vb_names=[key[:key.index(":")] for key in stat_hash.keys()]

            for name in vb_names:
                stat[name]=[stat_hash[name + ":state"], stat_hash[name+":counted"]]
            _nodes_stats[node.ip+":"+str(port)] = stat
        log.info(_nodes_stats)
        return _nodes_stats


    @staticmethod
    def pick_node(master):
        rest = RestConnection(master)
        nodes = rest.node_statuses()
        node_picked = None
        nodes_on_same_ip = True

        firstIp = nodes[0].ip
        for node in nodes:
            if node.ip != firstIp:
                nodes_on_same_ip = False
                break

        for node in nodes:
            node_picked = node
            if not nodes_on_same_ip:
                if node_picked.ip != master.ip:
                    log.info(
                        "Picked node ... {0}:{1}".format(node_picked.ip, node_picked.port))
                    break
            else:
                # temp fix - port numbers of master(machine ip and localhost: 9000 match
                if int(node_picked.port) == int(
                    master.port):
                    log.info("Not picking the master node {0}:{1}.. try again...".format(node_picked.ip,
                        node_picked.port))
                else:
                    log.info(
                        "Picked  node {0}:{1}".format(node_picked.ip, node_picked.port))
                    break
        return node_picked

    @staticmethod
    def pick_nodes(master, howmany=1):
        rest = RestConnection(master)
        nodes = rest.node_statuses()
        picked = []
        for node_for_stat in nodes:
            if node_for_stat.id.find('127.0.0.1') == -1 and node_for_stat.id.find(master.ip) == -1:
                picked.append(node_for_stat)
                if len(picked) == howmany:
                    break
        return picked



# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bits and pieces used by the driver that don't really fit elsewhere."""

try:
    import hashlib
    _md5func = hashlib.md5
except:  # for Python < 2.5
    import md5
    _md5func = md5.new
import random
import struct

import bson
from bson.son import SON
import pymongo
from pymongo.errors import (AutoReconnect,
                            OperationFailure,
                            TimeoutError)


def _index_list(key_or_list, direction=None):
    """Helper to generate a list of (key, direction) pairs.

    Takes such a list, or a single key, or a single key and direction.
    """
    if direction is not None:
        return [(key_or_list, direction)]
    else:
        if isinstance(key_or_list, basestring):
            return [(key_or_list, pymongo.ASCENDING)]
        elif not isinstance(key_or_list, list):
            raise TypeError("if no direction is specified, "
                            "key_or_list must be an instance of list")
        return key_or_list


def _index_document(index_list):
    """Helper to generate an index specifying document.

    Takes a list of (key, direction) pairs.
    """
    if isinstance(index_list, dict):
        raise TypeError("passing a dict to sort/create_index/hint is not "
                        "allowed - use a list of tuples instead. did you "
                        "mean %r?" % list(index_list.iteritems()))
    elif not isinstance(index_list, list):
        raise TypeError("must use a list of (key, direction) pairs, "
                        "not: " + repr(index_list))
    if not len(index_list):
        raise ValueError("key_or_list must not be the empty list")

    index = SON()
    for (key, value) in index_list:
        if not isinstance(key, basestring):
            raise TypeError("first item in each key pair must be a string")
        if value not in [pymongo.ASCENDING, pymongo.DESCENDING, pymongo.GEO2D, pymongo.GEOHAYSTACK]:
            raise TypeError("second item in each key pair must be ASCENDING, "
                            "DESCENDING, GEO2D, or GEOHAYSTACK")
        index[key] = value
    return index


def _unpack_response(response, cursor_id=None, as_class=dict, tz_aware=False):
    """Unpack a response from the database.

    Check the response for errors and unpack, returning a dictionary
    containing the response data.

    :Parameters:
      - `response`: byte string as returned from the database
      - `cursor_id` (optional): cursor_id we sent to get this response -
        used for raising an informative exception when we get cursor id not
        valid at server response
      - `as_class` (optional): class to use for resulting documents
    """
    response_flag = struct.unpack("<i", response[:4])[0]
    if response_flag & 1:
        # Shouldn't get this response if we aren't doing a getMore
        assert cursor_id is not None

        raise OperationFailure("cursor id '%s' not valid at server" %
                               cursor_id)
    elif response_flag & 2:
        error_object = bson.BSON(response[20:]).decode()
        if error_object["$err"].startswith("not master"):
            raise AutoReconnect("master has changed")
        raise OperationFailure("database error: %s" %
                               error_object["$err"])

    result = {}
    result["cursor_id"] = struct.unpack("<q", response[4:12])[0]
    result["starting_from"] = struct.unpack("<i", response[12:16])[0]
    result["number_returned"] = struct.unpack("<i", response[16:20])[0]
    result["data"] = bson.decode_all(response[20:], as_class, tz_aware)
    assert len(result["data"]) == result["number_returned"]
    return result


def _check_command_response(response, reset, msg="%s", allowable_errors=[]):
    if not response["ok"]:
        if "wtimeout" in response and response["wtimeout"]:
            raise TimeoutError(msg % response["errmsg"])
        if not response["errmsg"] in allowable_errors:
            if response["errmsg"] == "not master":
                if reset is not None:
                    reset()
                raise AutoReconnect("not master")
            if response["errmsg"] == "db assertion failure":
                ex_msg = ("db assertion failure, assertion: '%s'" %
                          response.get("assertion", ""))
                if "assertionCode" in response:
                    ex_msg += (", assertionCode: %d" %
                               (response["assertionCode"],))
                raise OperationFailure(ex_msg, response.get("assertionCode"))
            raise OperationFailure(msg % response["errmsg"])


def _password_digest(username, password):
    """Get a password digest to use for authentication.
    """
    if not isinstance(password, basestring):
        raise TypeError("password must be an instance of basestring")
    if not isinstance(username, basestring):
        raise TypeError("username must be an instance of basestring")

    md5hash = _md5func()
    md5hash.update("%s:mongo:%s" % (username.encode('utf-8'),
                                    password.encode('utf-8')))
    return unicode(md5hash.hexdigest())


def _auth_key(nonce, username, password):
    """Get an auth key to use for authentication.
    """
    digest = _password_digest(username, password)
    md5hash = _md5func()
    md5hash.update("%s%s%s" % (nonce, unicode(username), digest))
    return unicode(md5hash.hexdigest())


def _fields_list_to_dict(fields):
    """Takes a list of field names and returns a matching dictionary.

    ["a", "b"] becomes {"a": 1, "b": 1}

    and

    ["a.b.c", "d", "a.c"] becomes {"a.b.c": 1, "d": 1, "a.c": 1}
    """
    as_dict = {}
    for field in fields:
        if not isinstance(field, basestring):
            raise TypeError("fields must be a list of key names as "
                            "(string, unicode)")
        as_dict[field] = 1
    return as_dict

def shuffled(sequence):
    """Returns a copy of the sequence (as a :class:`list`) which has been
    shuffled by :func:`random.shuffle`.
    """
    out = list(sequence)
    random.shuffle(out)
    return out

